#!/usr/bin/python3
# coding=utf-8
#   python interface for dufoern usb stick
#   Copyright (C) 2017 Paul Görgen
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; in version 2 of the license
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA


import argparse
import logging
import re
import time
from cmd import Cmd

from pyduofern.duofern_stick import DuofernStickThreaded

parser = argparse.ArgumentParser(epilog="use at your own risk")

parser.add_argument('--configfile', help='location of system config file', default=None, metavar="CONFIGFILE")

parser.add_argument('--code', help='set 4-digit hex code for the system (warning, always use same code for once-paired '
                                   'devices) best chose something on the first run and write it down.',
                    default=None, metavar="SYSTEMCODE")

parser.add_argument('--device',
                    help='path to serial port created by duofern stick, defaults to first found serial port, typically '
                         'something like /dev/ttyUSB0 or /dev/duofernstick if you use the provided udev rules file',
                    default=None)

parser.add_argument('--pair', action='store_true',
                    help='Stick will wait for pairing for {} seconds. Afterwards look in the config file for the paired '
                         'device name or call this program again to list the newly found device. If devices that were '
                         'paired with this SYSTEMCODE are found while waiting for pairing, these are also added to'
                         'CONFIGFILE',
                    default=False)

parser.add_argument('--unpair', action='store_true',
                    help='Stick will wait for pairing for {} seconds. Afterwards look in the config file for the paired '
                         'device name or call this program again to list the newly found device. If devices that were '
                         'paired with this SYSTEMCODE are found while waiting for pairing, these are also added to'
                         'CONFIGFILE',
                    default=False)

parser.add_argument('--remote',
                    help='Pair by code. Added devices are also added to CONFIGFILE',
                    metavar='DEVICE_ID', default=None)

parser.add_argument('--pairtime', help='time to wait for pairing requests', metavar="SECONDS", default=60, type=int)

parser.add_argument('--refresh', action='store_true',
                    help='just sit and listen for devices that were already paired with the current system code but '
                         'were lost from the config file', default=False)
parser.add_argument('--refreshtime', help='time to spend refreshing',
                    metavar="SECONDS", default=60, type=int)

parser.add_argument('--set_name', help='Set name for a device.', nargs=2, default=None,
                    metavar=("DEVICE_ID", "DEVICE_NAME"))

parser.add_argument('--debug', help='enable verbose logging', action='store_true', default=False)

parser.add_argument('--up', help='pull up the selected rollershutter / blinds', metavar="NAME", nargs='+', default=None)
parser.add_argument('--down', help='roll down the selected rollershutter / blinds', metavar="NAME", nargs='+', default=None)
parser.add_argument('--stop', help='stop the selected rollershutter / blinds', metavar="NAME", nargs='+', default=None)

parser.add_argument('--on', help='switch on (for "Steckdosenaktor")', metavar="NAME", nargs='+', default=None)
parser.add_argument('--off', help='switch off (for "Steckdosenaktor")', metavar="NAME", nargs='+',
                    default=None)
parser.add_argument('--stairwell_on', help='switch on stairwell (for "Steckdosenaktor")', metavar="NAME", nargs='+',
                    default=None)
parser.add_argument('--stairwell_off', help='switch on stairwell (for "Steckdosenaktor")', metavar="NAME", nargs='+',
                    default=None)
parser.add_argument('--position', help='move shutter NAME to position POSITION', metavar=('POSITION', 'NAME'), nargs=2,
                    default=None, type=str)


def splitargs(func):
    def wrapper(*args, **kwargs):
        func(args[0], args[1].split(" "))

    return wrapper


def ids_for_names(func):
    def wrapper(*args, **kwargs):
        id_dict = {device['id']: device['name'] for device in args[0].stick.config['devices']
                   if device['name'] in args[1]}
        func(args[0], id_dict)

    return wrapper


class DuofernCLI(Cmd):
    def __init__(self, serial_port=None, system_code=None, config_file=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stick = DuofernStickThreaded(serial_port=args.device, system_code=args.code,
                                          config_file_json=args.configfile)
        self.stick._initialize()
        self.stick.start()
        self.prompt = "duofern> "

    def emptyline(self):
        pass

    def do_pair(self, args):
        """
        Usage:
          pair <TIMEOUT>
        
        Start pairing mode. Pass a timeout in seconds as <TIMEOUT>.
        Will return after the timeout if no devices start pairing within the given timeout.
        
        Example:
        duofern> pair 10
            
        """
        timeout = 10
        if len(args) != 0:
            try:
                timeout = int(args[0])
            except:
                print("Please use an integer number to indicate TIMEOUT in seconds")
        print("Starting pairing mode... waiting  {} seconds".format(int(timeout)))
        self.stick.pair(timeout=timeout)
        time.sleep(args.pairtime + 0.5)
        self.stick.sync_devices()
        print("Pairing done, Config file updated.")

    def do_unpair(self, args):
        """
        Usage:
          unpair <TIMEOUT>

        Start pairing mode. Pass a timeout in seconds as <TIMEOUT>.
        Will return after the timeout if no devices start pairing within the given timeout.

        Example:
        duofern> unpair 10

        """
        timeout = 10
        if len(args) != 0:
            try:
                timeout = int(args[0])
            except:
                print("Please use an integer number to indicate TIMEOUT in seconds")
        print("Starting pairing mode... waiting  {} seconds".format(int(timeout)))
        self.stick.unpair(timeout=timeout)
        time.sleep(args.pairtime + 0.5)
        self.stick.sync_devices()
        print("Pairing done, Config file updated.")

    def do_remote(self, args):
        code = args[0][0:6]
        timeout = int(args[1])
        self.stick.remote(code, timeout)
        time.sleep(args.pairtime + 0.5)
        self.stick.sync_devices()
        print("Pairing done, Config file updated.")

    @splitargs
    @ids_for_names
    def do_up(self, blinds):
        """
        Usage:
          up <SHUTTER> [<SHUTTER> <SHUTTER>]
          
        Lift one or several shutters. Accepts a list of shutter names sepatated by space.
        
        Example:
            duofern> up Livingroom
            duofern> up Livingroom Kitchen
        """
        for blind_id in blinds:
            print("lifting {}".format(blinds[blind_id]))
            self.stick.command(blind_id, "up")

    @splitargs
    @ids_for_names
    def do_down(self, blinds):
        """
        Usage:
          up <SHUTTER> [<SHUTTER> <SHUTTER>...]

        Lower one or several shutters. Accepts a list of shutter names sepatated by space.

        Example:
            duofern> up Livingroom
            duofern> up Livingroom Kitchen
        """
        for blind_id in blinds:
            print("lowering {}".format(blinds[blind_id]))
            self.stick.command(blind_id, "down")

    @splitargs
    def do_rename(self, args):
        """
        Usage:
          rename <NAME> <NEW_NAME>
        
        Rename an actor. Write changes to config file when done.
        
        Example:
            duofern> rename 13f897 kitchen_west
        """
        id = [device['id'] for device in self.stick.config['devices'] if device['name'] == args[0]]
        if len(id)==0:
            print("Please enter a valid device name for renaming.")
        self.stick.set_name(id[0], args[1])
        print("Set name for {} to {}".format(id[0], args[0]))

    def refresh(self,args):
        """
        Usage:
          refresh
        
        Refresh config file with current changes. 
        
        example:
            duofern> refresh
        """
        self.stick.sync_devices()

    @splitargs
    @ids_for_names
    def do_on(self, blinds):
        """
        Usage:
          off <SWITCH1> [<SWITCH2> <SWITCH3>]

        Switch on one or several switch actors. Accepts a list of actor names.

        Example:
            duofern> off Livingroom
            duofern> off Livingroom Kitchen
        """
        for blind_id in blinds:
            print("lifting {}".format(blinds[blind_id]))
            self.stick.command(blind_id, "up")

    @splitargs
    @ids_for_names
    def do_off(self, blinds):
        """
        Usage:
          off <SWITCH1> [<SWITCH2> <SWITCH3>]

        Switch off one or several switch actors. Accepts a list of actor names.

        Example:
            duofern> off Livingroom
            duofern> off Livingroom Kitchen
        """
        for blind_id in blinds:
            print("lifting {}".format(blinds[blind_id]))
            self.stick.command(blind_id, "up")


if __name__ == "__main__":
    args = parser.parse_args()

    # print(args.up)
    if args.debug:
        logging.basicConfig(format='    %(asctime)s: %(message)s', level=logging.DEBUG)
    else:
        logging.basicConfig(format='    %(asctime)s: %(message)s', level=logging.INFO)

    if args.code is not None:
        assert len(args.code) == 4, "System code must be a 4 digit hex code"
        try:
            bytearray.fromhex(args.code)
        except:
            print("System code must be a 4 digit hex code")
            exit(1)

    stick = DuofernStickThreaded(serial_port=args.device, system_code=args.code, config_file_json=args.configfile)

    if args.set_name is not None:
        assert len(args.set_name[0]) == 6 and re.match("^[0-9a-f]+$", args.set_name[0], re.IGNORECASE), \
            "id for renaming must be a valid 6 digit hex ID not {}".format(args.set_name[0])
        stick.set_name(args.set_name[0], args.set_name[1])
        stick.sync_devices()

    print("\nThe following devices are configured:\n")
    print("\n".join(
        ["id: {:6}    name: {}".format(device['id'], device['name']) for device in stick.config['devices']]))
    print("\n")

    if args.pair:
        print("entering pairing mode")
        stick._initialize()
        stick.start()
        stick.pair(timeout=args.pairtime)
        time.sleep(args.pairtime + 0.5)
        stick.sync_devices()
        print("""""")

    if args.unpair:
        print("entering pairing mode")
        stick._initialize()
        stick.start()
        stick.unpair(timeout=args.pairtime)
        time.sleep(args.pairtime + 0.5)
        stick.sync_devices()
        print("""""")

    if args.remote:
        print("entering remote pairing mode")
        stick._initialize()
        stick.start()
        stick.pair(timeout=args.pairtime)
        stick.remote(code=args.remote, timeout=args.pairtime)
        time.sleep(args.pairtime + 0.5)
        stick.sync_devices()
        print("""""")

    if args.refresh:
        stick._initialize()
        stick.start()
        time.sleep(args.refreshtime + 0.5)
        stick.sync_devices()

    if args.up:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.up]
        for blind_id in ids:
            stick.command(blind_id, "up")
            time.sleep(0.5)
            stick.command(blind_id, "up")
            time.sleep(0.5)
            stick.command(blind_id, "up")
            time.sleep(2)

    if args.down:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.down]
        for blind_id in ids:
            stick.command(blind_id, "down")
            time.sleep(0.5)
            stick.command(blind_id, "down")
            time.sleep(0.5)
            stick.command(blind_id, "down")
            time.sleep(2)
    
    if args.stop:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.stop]
        for blind_id in ids:
            stick.command(blind_id, "stop")
            time.sleep(0.5)
            stick.command(blind_id, "stop")
            time.sleep(0.5)
            stick.command(blind_id, "stop")
            time.sleep(2)

    if args.on:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.on]
        for blind_id in ids:
            stick.command(blind_id, "on")
            time.sleep(2)

    if args.off:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.off]
        for blind_id in ids:
            stick.command(blind_id, "off")
            time.sleep(2)

    if args.stairwell_on:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.stairwell_on]
        for blind_id in ids:
            stick.command(blind_id, "stairwellTime", 10)
            time.sleep(0.5)
            stick.command(blind_id, "stairwellFunction", "on")
            time.sleep(0.5)
            stick.command(blind_id, "on")
            time.sleep(2)

    if args.stairwell_off:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.stairwell_off]
        for blind_id in ids:
            stick.command(blind_id, "stairwellFunction", "off")
            time.sleep(0.5)
            stick.command(blind_id, "on")
            time.sleep(2)

    if args.position is not None:
        stick._initialize()
        stick.start()
        time.sleep(1)
        ids = [device['id'] for device in stick.config['devices'] if device['name'] in args.position[1:]]
        for blind_id in ids:
            stick.command(blind_id, "position", 100 - int(args.position[0]))
            time.sleep(0.5)
            stick.command(blind_id, "position", 100 - int(args.position[0]))
            time.sleep(0.5)
            stick.command(blind_id, "position", 100 - int(args.position[0]))
            time.sleep(2)

    stick.stop()
    time.sleep(1)
    try:
        stick.join()
    except RuntimeError:
        pass



# coding=utf-8
#   python interface for dufoern usb stick
#   Copyright (C) 2017 Paul Görgen
#   Rough python re-write of the FHEM duofern modules by telekatz, also licensed under GPLv2
#   This re-write contains only negligible amounts of original code
#   apart from some comments to facilitate translation of the not-yet
#   translated parts of the original software. Modification dates are
#   documented as submits to the git repository of this code, currently
#   maintained at https://github.com/gluap/pyduofern.git

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

devices = {
    "40": "RolloTron Standard",
    "41": "RolloTron Comfort Slave",
    "42": "Rohrmotor-Aktor",
    "43": "Universalaktor",
    "46": "Steckdosenaktor",
    "47": "Rohrmotor Steuerung",
    "48": "Dimmaktor",
    "49": "Rohrmotor",
    "4b": "Connect-Aktor",
    "4C": "Troll Basis",
    "4e": "SX5",
    "61": "RolloTron Comfort Master",
    "62": "Super Fake Device",
    "65": "Bewegungsmelder",
    "69": "Umweltsensor",
    "70": "Troll Comfort DuoFern",
    "71": "Troll Comfort DuoFern (Lichtmodus)",
    "73": "Raumthermostat",
    "74": "Wandtaster 6fach 230V",
    "a0": "Handsender (6 Gruppen-48 Geraete)",
    "a1": "Handsender (1 Gruppe-48 Geraete)",
    "a2": "Handsender (6 Gruppen-1 Geraet)",
    "a3": "Handsender (1 Gruppe-1 Geraet)",
    "a4": "Wandtaster",
    "a5": "Sonnensensor",
    "a7": "Funksender UP",
    "a8": "HomeTimer",
    "aa": "Markisenwaechter",
    "ab": "Rauchmelder",
    "ad": "Wandtaster 6fach Bat",
}

sensorMsg = {
    "0701": {"name": "up", "chan": 6, "state": "Btn01"},
    "0702": {"name": "stop", "chan": 6, "state": "Btn02"},
    "0703": {"name": "down", "chan": 6, "state": "Btn03"},
    "0718": {"name": "stepUp", "chan": 6, "state": "Btn18"},
    "0719": {"name": "stepDown", "chan": 6, "state": "Btn19"},
    "071a": {"name": "pressed", "chan": 6, "state": "Btn1A"},
    "0713": {"name": "dawn", "chan": 5, "state": "dawn"},
    "0709": {"name": "dusk", "chan": 5, "state": "dusk"},
    "0708": {"name": "startSun", "chan": 5, "state": "on"},
    "070a": {"name": "endSun", "chan": 5, "state": "off"},
    "070d": {"name": "startWind", "chan": 5, "state": "on"},
    "070e": {"name": "endWind", "chan": 5, "state": "off"},
    "0711": {"name": "startRain", "chan": 5, "state": "on"},
    "0712": {"name": "endRain", "chan": 5, "state": "off"},
    "071c": {"name": "startTemp", "chan": 5, "state": "on"},
    "071d": {"name": "endTemp", "chan": 5, "state": "off"},
    "071e": {"name": "startSmoke", "chan": 5, "state": "on"},
    "071f": {"name": "endSmoke", "chan": 5, "state": "off"},
    "0720": {"name": "startMotion", "chan": 5, "state": "on"},
    "0721": {"name": "endMotion", "chan": 5, "state": "off"},
    "0723": {"name": "closeEnd", "chan": 5, "state": "off"},
    "0724": {"name": "closeStart", "chan": 5, "state": "on"},
    "0e01": {"name": "off", "chan": 6, "state": "Btn01"},
    "0e02": {"name": "off", "chan": 6, "state": "Btn02"},
    "0e03": {"name": "on", "chan": 6, "state": "Btn03"},
}

deadTimes = {
    0x00: "off",
    0x10: "short(160ms)",
    0x20: "long(480ms)",
    0x30: "individual",
}

closingTimes = {
    0x00: "off",
    0x01: "30",
    0x02: "60",
    0x03: "90",
    0x04: "120",
    0x05: "150",
    0x06: "180",
    0x07: "210",
    0x08: "240",
    0x09: "error",
    0x0A: "error",
    0x0B: "error",
    0x0C: "error",
    0x0D: "error",
    0x0E: "error",
    0x0F: "error",
}

openSpeeds = {
    0x00: "error",
    0x10: "11",
    0x20: "15",
    0x30: "19",
}

commands = {
    "remotePair": {"noArg": "06010000000000000000"},
    "remoteUnpair": {"noArg": "06020000000000000000"},
    "up": {"noArg": "0701tt00000000000000"},
    "stop": {"noArg": "07020000000000000000"},
    "down": {"noArg": "0703tt00000000000000"},
    "position": {"value": "0707ttnn000000000000"},
    "level": {"value": "0707ttnn000000000000"},
    "sunMode": {"on": "070801FF000000000000",
                "off": "070A0100000000000000"},
    "dusk": {"noArg": "070901FF000000000000"},
    "reversal": {"noArg": "070C0000000000000000"},
    "modeChange": {"noArg": "070C0000000000000000"},
    "windMode": {"on": "070D01FF000000000000",
                 "off": "070E0100000000000000"},
    "rainMode": {"on": "071101FF000000000000",
                 "off": "07120100000000000000"},
    "dawn": {"noArg": "071301FF000000000000"},
    "rainDirection": {"down": "071400FD000000000000",
                      "up": "071400FE000000000000"},
    "windDirection": {"down": "071500FD000000000000",
                      "up": "071500FE000000000000"},
    "tempUp": {"noArg": "0718tt00000000000000"},
    "tempDown": {"noArg": "0719tt00000000000000"},
    "toggle": {"noArg": "071A0000000000000000"},
    "slatPosition": {"value": "071B00000000nn000000"},
    "desired-temp": {"temp1": "0722tt0000wwww000000"},
    "sunAutomatic": {"on": "080100FD000000000000",
                     "off": "080100FE000000000000"},
    "sunPosition": {"value": "080100nn000000000000"},
    "ventilatingMode": {"on": "080200FD000000000000",
                        "off": "080200FE000000000000"},
    "ventilatingPosition": {"value": "080200nn000000000000"},
    "intermediateMode": {"on": "080200FD000000000000",
                         "off": "080200FE000000000000"},
    "intermediateValue": {"value": "080200nn000000000000"},
    "saveIntermediateOnStop": {"on": "080200FB000000000000",
                               "off": "080200FC000000000000"},
    "runningTime": {"value3": "0803nn00000000000000"},
    "timeAutomatic": {"on": "080400FD000000000000",
                      "off": "080400FE000000000000"},
    "duskAutomatic": {"on": "080500FD000000000000",
                      "off": "080500FE000000000000"},
    "manualMode": {"on": "080600FD000000000000",
                   "off": "080600FE000000000000"},
    "windAutomatic": {"on": "080700FD000000000000",
                      "off": "080700FE000000000000"},
    "rainAutomatic": {"on": "080800FD000000000000",
                      "off": "080800FE000000000000"},
    "dawnAutomatic": {"on": "080900FD000000000000",
                      "off": "080900FE000000000000"},
    "tiltInSunPos": {"on": "080C00FD000000000000",
                     "off": "080C00FE000000000000"},
    "tiltInVentPos": {"on": "080D00FD000000000000",
                      "off": "080D00FE000000000000"},
    "tiltAfterMoveLevel": {"on": "080E00FD000000000000",
                           "off": "080E00FE000000000000"},
    "tiltAfterStopDown": {"on": "080F00FD000000000000",
                          "off": "080F00FE000000000000"},
    "defaultSlatPos": {"value": "0810nn00000000000000"},
    "blindsMode": {"on": "081100FD000000000000",
                   "off": "081100FE000000000000"},
    "slatRunTime": {"value4": "0812nn00000000000000"},
    "motorDeadTime": {"off": "08130000000000000000",
                      "short": "08130100000000000000",
                      "long": "08130200000000000000"},
    "stairwellFunction": {"on": "081400FD000000000000",
                          "off": "081400FE000000000000"},
    "stairwellTime": {"value2": "08140000wwww00000000"},
    "reset": {"settings": "0815CB00000000000000",
              "full": "0815CC00000000000000"},
    "10minuteAlarm": {"on": "081700FD000000000000",
                      "off": "081700FE000000000000"},
    "automaticClosing": {"off": "08180000000000000000",
                         "30": "08180001000000000000",
                         "60": "08180002000000000000",
                         "90": "08180003000000000000",
                         "120": "08180004000000000000",
                         "150": "08180005000000000000",
                         "180": "08180006000000000000",
                         "210": "08180007000000000000",
                         "240": "08180008000000000000"},
    "2000cycleAlarm": {"on": "081900FD000000000000",
                       "off": "081900FE000000000000"},
    "openSpeed": {"11": "081A0001000000000000",
                  "15": "081A0002000000000000",
                  "19": "081A0003000000000000"},
    "backJump": {"on": "081B00FD000000000000",
                 "off": "081B00FE000000000000"},
    "temperatureThreshold1": {"temp2": "081E00000001nn000000"},
    "temperatureThreshold2": {"temp2": "081E0000000200nn0000"},
    "temperatureThreshold3": {"temp2": "081E000000040000nn00"},
    "temperatureThreshold4": {"temp2": "081E00000008000000nn"},
    "actTempLimit": {"1": "081Ett00001000000000",
                     "2": "081Ett00003000000000",
                     "3": "081Ett00005000000000",
                     "4": "081Ett00007000000000"},
    "on": {"noArg": "0E03tt00000000000000"},
    "off": {"noArg": "0E02tt00000000000000"},
}

wCmds = {
    "interval": {"enable": 0x80, "min": 1, "max": 100, "offset": 0,
                 "reg": 7, "byte": 0, "size": 1, "count": 1,
                 "mask": 0xff, "shift": 0},
    "DCF": {"enable": 0x02, "min": 0, "max": 0, "offset": 0,
            "reg": 7, "byte": 1, "size": 1, "count": 1,
            "mask": 0x02, "shift": 0},
    "timezone": {"enable": 0x00, "min": 0, "max": 23, "offset": 0,
                 "reg": 7, "byte": 4, "size": 1, "count": 1,
                 "mask": 0xff, "shift": 0},
    "latitude": {"enable": 0x00, "min": 0, "max": 90, "offset": 0,
                 "reg": 7, "byte": 5, "size": 1, "count": 1,
                 "mask": 0xff, "shift": 0},
    "longitude": {"enable": 0x00, "min": -90, "max": 90, "offset": 256,
                  "reg": 7, "byte": 7, "size": 1, "count": 1,
                  "mask": 0xff, "shift": 0},
    "triggerWind": {"enable": 0x20, "min": 1, "max": 31, "offset": 0,
                    "reg": 6, "byte": 0, "size": 1, "count": 5,
                    "mask": 0x7f, "shift": 0},
    "triggerRain": {"enable": 0x80, "min": 0, "max": 0, "offset": 0,
                    "reg": 6, "byte": 0, "size": 1, "count": 1,
                    "mask": 0x80, "shift": 0},
    "triggerTemperature": {"enable": 0x80, "min": -40, "max": 80, "offset": 40,
                           "reg": 6, "byte": 5, "size": 1, "count": 5,
                           "mask": 0xff, "shift": 0},
    "triggerDawn": {"enable": 0x10000000, "min": 1, "max": 100, "offset": -1,
                    "reg": 0, "byte": 0, "size": 4, "count": 5,
                    "mask": 0x1000007F, "shift": 0},
    "triggerDusk": {"enable": 0x20000000, "min": 1, "max": 100, "offset": -1,
                    "reg": 0, "byte": 0, "size": 4, "count": 5,
                    "mask": 0x201FC000, "shift": 14},
    "triggerSun": {"enable": 0x20000000, "min": 1, "max": 0x3FFFFFFF, "offset": 0,
                   "reg": 3, "byte": 0, "size": 4, "count": 5,
                   "mask": 0x3FFFFFC0, "shift": 0},
    "triggerSunDirection": {"enable": 0x00, "min": 1, "max": 0xFF, "offset": 0,
                            "reg": 3, "byte": 1, "size": 4, "count": 5,
                            "mask": 0x000000FF, "shift": 0},
    "triggerSunHeight": {"enable": 0x00, "min": 1, "max": 0x1FFF, "offset": 0,
                         "reg": 3, "byte": 1, "size": 4, "count": 5,
                         "mask": 0x00001F80, "shift": 0},
}

commandsStatus = {
    "getStatus": "0F",
    "getWeather": "13",
    "getTime": "10",
}

setsBasic = {
    "reset:settings,full": "",
    "remotePair:noArg": "",
    "remoteUnpair:noArg": "",
}

setsDefaultRollerShutter = {
    "getStatus:noArg": "",
    "up:noArg": "",
    "down:noArg": "",
    "stop:noArg": "",
    "toggle:noArg": "",
    "dusk:noArg": "",
    "dawn:noArg": "",
    "sunMode:on,off": "",
    "position:slider,0,1,100": "",
    "sunPosition:slider,0,1,100": "",
    "ventilatingPosition:slider,0,1,100": "",
    "dawnAutomatic:on,off": "",
    "duskAutomatic:on,off": "",
    "manualMode:on,off": "",
    "sunAutomatic:on,off": "",
    "timeAutomatic:on,off": "",
    "ventilatingMode:on,off": "",
}

setsRolloTube = {
    "windAutomatic:on,off": "",
    "rainAutomatic:on,off": "",
    "windDirection:up,down": "",
    "rainDirection:up,down": "",
    "windMode:on,off": "",
    "rainMode:on,off": "",
    "reversal:on,off": "",
}

setsTroll = {
    "windAutomatic:on,off": "",
    "rainAutomatic:on,off": "",
    "windDirection:up,down": "",
    "rainDirection:up,down": "",
    "windMode:on,off": "",
    "rainMode:on,off": "",
    "runningTime:slider,0,1,150": "",
    "motorDeadTime:off,short,long": "",
    "reversal:on,off": "",
}

setsBlinds = {
    "tiltInSunPos:on,off": "",
    "tiltInVentPos:on,off": "",
    "tiltAfterMoveLevel:on,off": "",
    "tiltAfterStopDown:on,off": "",
    "defaultSlatPos:slider,0,1,100": "",
    "slatRunTime:slider,0,100,5000": "",
    "slatPosition:slider,0,1,100": "",
}

setsSwitchActor = {
    "getStatus:noArg": "",
    "dawnAutomatic:on,off": "",
    "duskAutomatic:on,off": "",
    "manualMode:on,off": "",
    "sunAutomatic:on,off": "",
    "timeAutomatic:on,off": "",
    "sunMode:on,off": "",
    "modeChange:on,off": "",
    "stairwellFunction:on,off": "",
    "stairwellTime:slider,0,10,3200": "",
    "on:noArg": "",
    "off:noArg": "",
    "dusk:noArg": "",
    "dawn:noArg": "",
}

setsUmweltsensor = {
    "getStatus:noArg": "",
    "getWeather:noArg": "",
    "getTime:noArg": "",
}

setsUmweltsensor00 = {
    "getWeather:noArg": "",
    "getTime:noArg": "",
    "getConfig:noArg": "",
    "writeConfig:noArg": "",
    "DCF:on,off": "",
    "interval:off,1,2,3,4,5,6,7,8,9,10,15,20,30,40,50,60,70,80,90,100": "",
    "latitude": "",
    "longitude": "",
    "timezone": "",
    "time:noArg": "",
    "triggerDawn": "",
    "triggerDusk": "",
    "triggerRain:on,off": "",
    "triggerSun": "",
    "triggerSunDirection": "",
    "triggerSunHeight": "",
    "triggerTemperature": "",
    "triggerWind": "",
}

setsUmweltsensor01 = {
    "windAutomatic:on,off": "",
    "rainAutomatic:on,off": "",
    "windDirection:up,down": "",
    "rainDirection:up,down": "",
    "windMode:on,off": "",
    "rainMode:on,off": "",
    "runningTime:slider,0,1,100": "",
    "reversal:on,off": "",
}

setsSX5 = {
    "getStatus:noArg": "",
    "up:noArg": "",
    "down:noArg": "",
    "stop:noArg": "",
    "position:slider,0,1,100": "",
    "ventilatingPosition:slider,0,1,100": "",
    "manualMode:on,off": "",
    "timeAutomatic:on,off": "",
    "ventilatingMode:on,off": "",
    "10minuteAlarm:on,off": "",
    "automaticClosing:off,30,60,90,120,150,180,210,240": "",
    "2000cycleAlarm:on,off": "",
    "openSpeed:11,15,19": "",
    "backJump:on,off": "",
}

setsDimmer = {
    "getStatus:noArg": "",
    "level:slider,0,1,100": "",
    "on:noArg": "",
    "off:noArg": "",
    "dawnAutomatic:on,off": "",
    "duskAutomatic:on,off": "",
    "manualMode:on,off": "",
    "sunAutomatic:on,off": "",
    "timeAutomatic:on,off": "",
    "sunMode:on,off": "",
    "modeChange:on,off": "",
    "stairwellFunction:on,off": "",
    "stairwellTime:slider,0,10,3200": "",
    "runningTime:slider,0,1,255": "",
    "intermediateMode:on,off": "",
    "intermediateValue:slider,0,1,100": "",
    "saveIntermediateOnStop:on,off": "",
    "dusk:noArg": "",
    "dawn:noArg": "",
}

tempSetList = "4.0,4.5,5.0,5.5,6.0,6.5,7.0,7.5,8.0,8.5,9.0,9.5,10.0,10.5,11.0,11.5,12.0,12.5,13.0,13.5,14.0,14.5,15.0,15.5,16.0,16.5,17.0,17.5,18.0,18.5,19.0,19.5,20.0,20.5,21.0,21.5,22.0,22.5,23.0,23.5,24.0,24.5,25.0,25.5,26.0,26.5,27.0,27.5,28.0,28.5,29.0,29.5,30.0"

setsThermostat = {
    "getStatus:noArg": "",
    "tempUp:noArg": "",
    "tempDown:noArg": "",
    "manualMode:on,off": "",
    "timeAutomatic:on,off": "",
    "temperatureThreshold1:$tempSetList": "",
    "temperatureThreshold2:$tempSetList": "",
    "temperatureThreshold3:$tempSetList": "",
    "temperatureThreshold4:$tempSetList": "",
    "actTempLimit:0,1,2,3": "",
    "desired-temp:$tempSetList": "",
}



# coding=utf-8
#   python interface for dufoern usb stick
#   Copyright (C) 2017 Paul Görgen

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

class DuofernException(Exception):
    pass


class DuofernTimeoutException(DuofernException):
    pass



# coding=utf-8
#   python interface for dufoern usb stick
#   Copyright (C) 2017 Paul Görgen
#   Rough python re-write of the FHEM duofern modules by telekatz, also licensed under GPLv2
#   This re-write contains only negligible amounts of original code
#   apart from some comments to facilitate translation of the not-yet
#   translated parts of the original software. Modification dates are
#   documented as submits to the git repository of this code, currently
#   maintained at https://github.com/gluap/pyduofern.git

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import asyncio
import logging
import time

from .definitions import *

# regexe for replacing:
# hash->\{([^\}]+)\}\{([^\}]+)\}
# hash['$1']['$2']
#
# ^([^\n]+=) \(([^?\n]+)\?([^:\n]+):([^\)\n]+)\)?#?
# $1 $3 if $2 else $4

logger = logging.getLogger(__file__)

duoStatusRequest = "0DFFnn400000000000000000000000000000yyyyyy01"
duoCommand = "0Dkknnnnnnnnnnnnnnnnnnnn000000zzzzzzyyyyyy00"
duoWeatherConfig = "0D001B400000000000000000000000000000yyyyyy00"
duoWeatherWriteConfig = "0DFF1Brrnnnnnnnnnnnnnnnnnnnn00000000yyyyyy00"
duoSetTime = "0D0110800001mmmmmmmmnnnnnn0000000000yyyyyy00"


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def DoTrigger(*args):
    logger.debug("called DoTrigger({})".format(args))


#def readingsBulkUpdate(*args):
#    pass


#def readingsSingleUpdate(*args):
#    pass


#def  readingsEndUpdate(*args):
#    pass


#def  readingsBeginUpdate(*args):
#    pass


def RemoveInternalTimer(*args):
    pass


def DUOFERN_DecodeWeatherSensorConfig(*args):
    pass


class Duofern(object):
    def __init__(self, send_hook=None, asyncio=False, changes_callback=None):
        self.asyncio = asyncio
        self.modules = {'by_code': {}}  # i guess this is supposed to be a hash of self.modules...
        self.ignore_devices = {}  # should replace attrValrel
        assert send_hook is not None, "Must define send callback"
        self.send_hook = send_hook
        self.changes_callback = changes_callback
        pass

    def add_device(self, code, name=None):
        if name is None:
            name = len(self.modules['by_code'])
        logger.debug("adding {}".format(code))
        self.modules['by_code'][code] = {'name': name, 'channels': {None}}

    def del_device(self, code, name=None):
        if name is None:
            name = len(self.modules['by_code'])
        logger.info("removing {}".format(code))
        if code in self.modules['by_code']:
            del self.modules['by_code'][code]

    def update_state(self, code, key, value, trigger=None, channel: int = None):
        """

        :param code: duofern system code
        :param key: some arbitrary key that should be set in the state dict
        :param value: the corresponding value
        :param trigger: whether or not to call the callback
        :param channel: if this is a multichannel actor: The channel the key should be set for
        :return:
        """
        if channel is not None:
            channel_str = "{:02x}".format(channel)
            key = key + "_" + channel_str
            self.modules['by_code'][code]['channels'].add(channel_str)

        self.modules['by_code'][code][key] = value

        if self.changes_callback and trigger:
            self.changes_callback(code, key, value)

    def delete_state(self, code, key, channel: int = None):
        if channel is not None:
            channel_str = "{:02x}".format(channel)
            key = key + "_" + channel_str
        if key in self.modules['by_code'][code]:
            del self.modules['by_code'][code][key]

    def get_state(self, code, key, channel=None, default=None):
        if channel is not None:
            channel_str = "{:02x}".format(channel)
            key = key + "_" + channel_str

        if not key in self.modules['by_code'][code]:
            return default

        return self.modules['by_code'][code][key]

    def parse(self, msg):
        code = msg[30:36]
        if msg[0:2] == '81':
            code = msg[36:42]

        if code.lower() == 'ffffff':
            return
        # return hash->{NAME} if (code == "FFFFFF")

        try:
            # module_definition = self.modules['by_code'][code]
            name = self.modules['by_code'][code]['name']

        except KeyError:
            self.add_device(code)
            logger.info("detected unknown device, ID={}".format(code))
            name = self.modules['by_code'][code]['name']

        #hash="asdf"
        # module_definition01 = None
        # module_definition02 = None
        channel2 = None

        #        if not module_definition:
        #            DoTrigger("global", "Undefined code {}".format(code))
        #            # module_definition = self.modules['by_code']{code}
        #            logger.warning("Undefined code {}".format(code))
        #            raise DuofernException("Undefined code {}".format(code))

        # hash = module_definition
        # name = hash['name']
        channel = None

        if name in self.ignore_devices:
            return name

        # Device paired
        if msg[0:4] == "0602":
            self.update_state(code, "state", "paired", "1", channel=channel)
            # del hash['READINGS']['unpaired']
            logger.info("DUOFERN device paired, ID {}".format(code))

        # Device unpaired
        elif (msg[0:4] == "0603"):
            # readingsBeginUpdate(hash)
            self.update_state(code, "unpaired", 1, "1", channel=channel)
            self.update_state(code, "state", "unpaired", "1", channel=channel)
            self.del_device(code)
            # # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch
            logger.warning("DUOFERN device unpaired, code {}".format(code))

        # Status Nachricht Aktor
        elif msg[0:6] == "0fff0f":
            format = msg[6:6 + 2]
            ver = msg[24:24 + 1] + msg[25:25 + 1]

            self.update_state(code, "version", ver, "0", channel=channel)

            # RemoveInternalTimer(hash)
            # del hash['helper']['timeout']

            # Bewegungsmelder, Wettersensor, Mehrfachwandtaster not tested yet
            if code[0:2] in ("65", "69", "74"):  # pragma: no cover
                #self.update_state(code, "state", "OK", "1", channel=channel)
                #module_definition01 = self.modules['by_code'][code + "01"]
                channel = 1
                #if not module_definition01:
                    #DoTrigger("global", "UNDEFINED DUOFERN_code_actor DUOFERN code01")
                    #module_definition01 = self.modules['by_code'][code + "01"]

            # Universalaktor -- not tested yet
            elif code[0:2] == "43":  # pragma: no cover
                self.update_state(code, "state", "OK", "1", channel=channel)
                #module_definition01 = self.modules['by_code'][code]
                channel = 1
                #if not module_definition01:
                #    DoTrigger("global", "UNDEFINED DUOFERN_code+_01 DUOFERN code+01")

                #module_definition02 = None
                channel2 = 2

            #if module_definition01:
                # it seems that sometimes "module_definition01" corresponts to channel "01", at other times
                # channel="00". I am trying to stick with what module_definition was set to.
                #hash = module_definition01
                #channel = 1

            # RolloTron
            if format == "21":
                pos = int(msg[22:22 + 2], 16) & 0x7F
                ventPos = int(msg[12:12 + 2], 16) & 0x7F
                ventMode = "on" if int(msg[12:12 + 2], 16) & 0x80 else "off"
                sunPos = int(msg[20:20 + 2], 16) & 0x7F
                sunMode = "on" if int(msg[20:20 + 2], 16) & 0x80 else "off"
                timerAuto = "on" if int(msg[8:8 + 2], 16) & 0x01 else "off"
                sunAuto = "on" if int(msg[8:8 + 2], 16) & 0x04 else "off"
                dawnAuto = "on" if int(msg[10:10 + 2], 16) & 0x08 else "off"
                duskAuto = "on" if int(msg[8:8 + 2], 16) & 0x08 else "off"
                manualMode = "on" if int(msg[8:8 + 2], 16) & 0x80 else "off"

                state = pos
                state = "opened" if (pos == 0) else pos
                state = "closed" if (pos == 100) else pos

                # readingsBeginUpdate(hash)
                self.update_state(code, "ventilatingPosition", ventPos, "1", channel=channel)
                self.update_state(code, "ventilatingMode", ventMode, "1", channel=channel)
                self.update_state(code, "sunPosition", sunPos, "1", channel=channel)
                self.update_state(code, "sunMode", sunMode, "1", channel=channel)
                self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel)
                self.update_state(code, "sunAutomatic", sunAuto, "1", channel=channel)
                self.update_state(code, "dawnAutomatic", dawnAuto, "1", channel=channel)
                self.update_state(code, "duskAutomatic", duskAuto, "1", channel=channel)
                self.update_state(code, "manualMode", manualMode, "1", channel=channel)
                self.update_state(code, "position", pos, "1", channel=channel)
                self.update_state(code, "state", state, "1", channel=channel)
                self.update_state(code, "moving", "stop", "1", channel=channel)
                # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch

            # Universal Aktor, Steckdosenaktor, Troll Comfort DuoFern (Lichtmodus) not tested yet
            elif format == "22":  # pragma: no cover
                level = int(msg[22:22 + 2], 16) & 0x7F
                modeChange = "on" if int(msg[22:22 + 2], 16) & 0x80 else "off"
                sunMode = "on" if int(msg[14:14 + 2], 16) & 0x10 else "off"
                timerAuto = "on" if int(msg[14:14 + 2], 16) & 0x01 else "off"
                sunAuto = "on" if int(msg[14:14 + 2], 16) & 0x04 else "off"
                dawnAuto = "on" if int(msg[14:14 + 2], 16) & 0x40 else "off"
                duskAuto = "on" if int(msg[14:14 + 2], 16) & 0x02 else "off"
                manualMode = "on" if int(msg[14:14 + 2], 16) & 0x20 else "off"
                stairwellFunction = "on" if int(msg[16:16 + 4], 16) & 0x8000 else "off"
                stairwellTime = (int(msg[16:16 + 4], 16) & 0x7FFF) / 10

                state = level
                if level == 0:
                    state = "off"
                if level == 100:
                    state = "on"

                # readingsBeginUpdate(hash)
                self.update_state(code, "sunMode", sunMode, "1", channel=channel)
                self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel)
                self.update_state(code, "sunAutomatic", sunAuto, "1", channel=channel)
                self.update_state(code, "dawnAutomatic", dawnAuto, "1", channel=channel)
                self.update_state(code, "duskAutomatic", duskAuto, "1", channel=channel)
                self.update_state(code, "manualMode", manualMode, "1", channel=channel)
                self.update_state(code, "modeChange", modeChange, "1", channel=channel)
                self.update_state(code, "stairwellFunction", stairwellFunction, "1", channel=channel)
                self.update_state(code, "stairwellTime", stairwellTime, "1", channel=channel)
                self.update_state(code, "level", level, "1", channel=channel)
                self.update_state(code, "state", state, "1", channel=channel)
                # readingsEndUpdate(hash, 1)

                if channel2 is not None:
                    level = int(msg[20:20 + 2], 16) & 0x7F
                    modeChange = "on" if int(msg[20:20 + 2], 16) & 0x80 else "off"
                    sunMode = "on" if int(msg[12:12 + 2], 16) & 0x10 else "off"
                    timerAuto = "on" if int(msg[12:12 + 2], 16) & 0x01 else "off"
                    sunAuto = "on" if int(msg[12:12 + 2], 16) & 0x04 else "off"
                    dawnAuto = "on" if int(msg[12:12 + 2], 16) & 0x40 else "off"
                    duskAuto = "on" if int(msg[12:12 + 2], 16) & 0x02 else "off"
                    manualMode = "on" if int(msg[12:12 + 2], 16) & 0x20 else "off"
                    stairwellFunction = "on" if int(msg[8:8 + 4], 16) & 0x8000 else "off"
                    stairwellTime = (int(msg[8:8 + 4], 16) & 0x7FFF) / 10

                    state = level
                    if level == 0:
                        state = "off"
                    if level == 100:
                        state = "on"

                    # readingsBeginUpdate(hash)
                    self.update_state(code, "sunMode", sunMode, "1", channel=channel2)
                    self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel2)
                    self.update_state(code, "sunAutomatic", sunAuto, "1", channel=channel2)
                    self.update_state(code, "dawnAutomatic", dawnAuto, "1", channel=channel2)
                    self.update_state(code, "duskAutomatic", duskAuto, "1", channel=channel2)
                    self.update_state(code, "manualMode", manualMode, "1", channel=channel2)
                    self.update_state(code, "modeChange", modeChange, "1", channel=channel2)
                    self.update_state(code, "stairwellFunction", stairwellFunction, "1", channel=channel2)
                    self.update_state(code, "stairwellTime", stairwellTime, "1", channel=channel2)
                    self.update_state(code, "level", level, "1", channel=channel2)
                    self.update_state(code, "state", state, "1", channel=channel2)
                    # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch
            elif format == "23":
                pos = int(msg[22:22 + 2], 16) & 0x7F
                reversal = "on" if int(msg[22:22 + 2], 16) & 0x80 else "off"
                ventPos = int(msg[16:16 + 2], 16) & 0x7F
                ventMode = "on" if int(msg[16:16 + 2], 16) & 0x80 else "off"
                sunPos = int(msg[18:18 + 2], 16) & 0x7F
                sunMode = "on" if int(msg[14:14 + 2], 16) & 0x10 else "off"
                timerAuto = "on" if int(msg[14:14 + 2], 16) & 0x01 else "off"
                sunAuto = "on" if int(msg[14:14 + 2], 16) & 0x04 else "off"
                dawnAuto = "on" if int(msg[12:12 + 2], 16) & 0x02 else "off"
                duskAuto = "on" if int(msg[14:14 + 2], 16) & 0x02 else "off"
                manualMode = "on" if int(msg[14:14 + 2], 16) & 0x20 else "off"
                windAuto = "on" if int(msg[14:14 + 2], 16) & 0x40 else "off"
                windMode = "on" if int(msg[14:14 + 2], 16) & 0x08 else "off"
                windDir = "down" if int(msg[12:12 + 2], 16) & 0x04 else "up"
                rainAuto = "on" if int(msg[14:14 + 2], 16) & 0x80 else "off"
                rainMode = "on" if int(msg[12:12 + 2], 16) & 0x01 else "off"
                rainDir = "down" if int(msg[12:12 + 2], 16) & 0x08 else "up"
                runningTime = int(msg[20:20 + 2], 16)
                deadTime = int(msg[12:12 + 2], 16) & 0x30
                blindsMode = "on" if int(msg[26:26 + 2], 16) & 0x80 else "off"
                tiltInSunPos = "on" if int(msg[18:18 + 2], 16) & 0x80 else "off"
                tiltInVentPos = "on" if int(msg[8:8 + 2], 16) & 0x80 else "off"
                tiltAfterMoveLevel = "on" if int(msg[8:8 + 2], 16) & 0x40 else "off"
                tiltAfterStopDown = "on" if int(msg[10:10 + 2], 16) & 0x80 else "off"
                defaultSlatPos = int(msg[10:10 + 2], 16) & 0x7F
                slatRunTime = int(msg[8:8 + 2], 16) & 0x3F
                slatPosition = int(msg[26:26 + 2], 16) & 0x7F

                state = "opened" if (pos == 0) else pos
                state = "closed" if (pos == 100) else state

                # readingsBeginUpdate(hash)
                self.update_state(code, "ventilatingPosition", ventPos, "1", channel=channel)
                self.update_state(code, "ventilatingMode", ventMode, "1", channel=channel)
                self.update_state(code, "sunPosition", sunPos, "1", channel=channel)
                self.update_state(code, "sunMode", sunMode, "1", channel=channel)
                self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel)
                self.update_state(code, "sunAutomatic", sunAuto, "1", channel=channel)
                self.update_state(code, "dawnAutomatic", dawnAuto, "1", channel=channel)
                self.update_state(code, "duskAutomatic", duskAuto, "1", channel=channel)
                self.update_state(code, "manualMode", manualMode, "1", channel=channel)
                self.update_state(code, "windAutomatic", windAuto, "1", channel=channel)
                self.update_state(code, "windMode", windMode, "1", channel=channel)
                self.update_state(code, "windDirection", windDir, "1", channel=channel)
                self.update_state(code, "rainAutomatic", rainAuto, "1", channel=channel)
                self.update_state(code, "rainMode", rainMode, "1", channel=channel)
                self.update_state(code, "rainDirection", rainDir, "1", channel=channel)
                self.update_state(code, "runningTime", runningTime, "1", channel=channel)
                self.update_state(code, "motorDeadTime", deadTimes[deadTime], "1", channel=channel)
                self.update_state(code, "position", pos, "1", channel=channel)
                self.update_state(code, "reversal", reversal, "1", channel=channel)
                self.update_state(code, "blindsMode", blindsMode, "1", channel=channel)

                # not tested yet
                if blindsMode == "on":  # pragma: no cover
                    self.update_state(code, "tiltInSunPos", tiltInSunPos, "1", channel=channel)
                    self.update_state(code, "tiltInVentPos", tiltInVentPos, "1", channel=channel)
                    self.update_state(code, "tiltAfterMoveLevel", tiltAfterMoveLevel, "1", channel=channel)
                    self.update_state(code, "tiltAfterStopDown", tiltAfterStopDown, "1", channel=channel)
                    self.update_state(code, "defaultSlatPos", defaultSlatPos, "1", channel=channel)
                    self.update_state(code, "slatRunTime", slatRunTime, "1", channel=channel)
                    self.update_state(code, "slatPosition", slatPosition, "1", channel=channel)
                else:
                    self.delete_state(code, 'tiltInSunPos', channel=channel)
                    self.delete_state(code, 'tiltInVentPos', channel=channel)
                    self.delete_state(code, 'tiltAfterMoveLevel', channel=channel)
                    self.delete_state(code, 'tiltAfterStopDown', channel=channel)
                    self.delete_state(code, 'defaultSlatPos', channel=channel)
                    self.delete_state(code, 'slatRunTime', channel=channel)
                    self.delete_state(code, 'slatPosition', channel=channel)

                self.update_state(code, "moving", "stop", "1", channel=channel)
                self.update_state(code, "state", state, "1", channel=channel)
                # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch
            # Rohrmotor, SX5  -- not tested yet
            elif format == "24":  # pragma: no cover

                pos = int(msg[22:22 + 2], 16) & 0x7F
                reversal = "on" if int(msg[22:22 + 2], 16) & 0x80 else "off"
                ventPos = int(msg[16:16 + 2], 16) & 0x7F
                ventMode = "on" if int(msg[16:16 + 2], 16) & 0x80 else "off"
                sunPos = int(msg[18:18 + 2], 16) & 0x7F
                sunMode = "on" if int(msg[14:14 + 2], 16) & 0x10 else "off"
                timerAuto = "on" if int(msg[14:14 + 2], 16) & 0x01 else "off"
                sunAuto = "on" if int(msg[14:14 + 2], 16) & 0x04 else "off"
                dawnAuto = "on" if int(msg[12:12 + 2], 16) & 0x02 else "off"
                duskAuto = "on" if int(msg[14:14 + 2], 16) & 0x02 else "off"
                manualMode = "on" if int(msg[14:14 + 2], 16) & 0x20 else "off"
                windAuto = "on" if int(msg[14:14 + 2], 16) & 0x40 else "off"
                windMode = "on" if int(msg[14:14 + 2], 16) & 0x08 else "off"
                windDir = "down" if int(msg[12:12 + 2], 16) & 0x04 else "up"
                rainAuto = "on" if int(msg[14:14 + 2], 16) & 0x80 else "off"
                rainMode = "on" if int(msg[12:12 + 2], 16) & 0x01 else "off"
                rainDir = "down" if int(msg[12:12 + 2], 16) & 0x08 else "up"
                obstacle = "1" if int(msg[12:12 + 2], 16) & 0x10 else "0"
                block = "1" if int(msg[12:12 + 2], 16) & 0x40 else "0"
                lightCurtain = "1" if int(msg[8:8 + 2], 16) & 0x80 else "0"
                autoClose = int(msg[10:10 + 2], 16) & 0x0F
                openSpeed = int(msg[10:10 + 2], 16) & 0x30
                alert2000 = "on" if int(msg[10:10 + 2], 16) & 0x80 else "off"
                backJump = "on" if int(msg[26:26 + 2], 16) & 0x01 else "off"
                alert10 = "on" if int(msg[26:26 + 2], 16) & 0x02 else "off"

                state = pos
                state = "opened" if (pos == 0) else pos
                state = "closed" if (pos == 100) else pos
                state = "light curtain" if (lightCurtain == "1") else pos
                state = "obstacle" if (obstacle == "1") else pos
                state = "block" if (block == "1") else pos

                # readingsBeginUpdate(hash)
                self.update_state(code, "manualMode", manualMode, "1", channel=channel)
                self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel)
                self.update_state(code, "ventilatingPosition", ventPos, "1", channel=channel)
                self.update_state(code, "ventilatingMode", ventMode, "1", channel=channel)
                self.update_state(code, "position", pos, "1", channel=channel)
                self.update_state(code, "state", state, "1", channel=channel)
                self.update_state(code, "obstacle", obstacle, "1", channel=channel)
                self.update_state(code, "block", block, "1", channel=channel)
                self.update_state(code, "moving", "stop", "1", channel=channel)

                if code[0:2] == "4E":  # SX5
                    self.update_state(code, "10minuteAlarm", alert10, "1", channel=channel)
                    self.update_state(code, "automaticClosing", closingTimes['autoClose'], "1", channel=channel)
                    self.update_state(code, "2000cycleAlarm", alert2000, "1", channel=channel)
                    self.update_state(code, "openSpeed", openSpeeds['openSpeed'], "1", channel=channel)
                    self.update_state(code, "backJump", backJump, "1", channel=channel)
                    self.update_state(code, "lightCurtain", lightCurtain, "1", channel=channel)
                else:
                    self.update_state(code, "sunPosition", sunPos, "1", channel=channel)
                    self.update_state(code, "sunMode", sunMode, "1", channel=channel)
                    self.update_state(code, "sunAutomatic", sunAuto, "1", channel=channel)
                    self.update_state(code, "dawnAutomatic", dawnAuto, "1", channel=channel)
                    self.update_state(code, "duskAutomatic", duskAuto, "1", channel=channel)
                    self.update_state(code, "windAutomatic", windAuto, "1", channel=channel)
                    self.update_state(code, "windMode", windMode, "1", channel=channel)
                    self.update_state(code, "windDirection", windDir, "1", channel=channel)
                    self.update_state(code, "rainAutomatic", rainAuto, "1", channel=channel)
                    self.update_state(code, "rainMode", rainMode, "1", channel=channel)
                    self.update_state(code, "rainDirection", rainDir, "1", channel=channel)
                    self.update_state(code, "reversal", reversal, "1", channel=channel)

                # readingsEndUpdate(hash, 1)

                # Dimmaktor -- not tested yet
            elif format == "25":  # pragma: no cover
                stairwellFunction = "on" if int(msg[10:10 + 4], 16) & 0x8000 else "off"
                stairwellTime = (int(msg[10:10 + 4], 16) & 0x7FFF) / 10
                timerAuto = "on" if int(msg[14:14 + 2], 16) & 0x01 else "off"
                duskAuto = "on" if int(msg[14:14 + 2], 16) & 0x02 else "off"
                sunAuto = "on" if int(msg[14:14 + 2], 16) & 0x04 else "off"
                sunMode = "on" if int(msg[14:14 + 2], 16) & 0x08 else "off"
                manualMode = "on" if int(msg[14:14 + 2], 16) & 0x20 else "off"
                dawnAuto = "on" if int(msg[14:14 + 2], 16) & 0x40 else "off"
                intemedSave = "on" if int(msg[14:14 + 2], 16) & 0x80 else "off"
                runningTime = int(msg[18:18 + 2], 16)
                intemedVal = int(msg[20:20 + 2], 16) & 0x7F
                intermedMode = "on" if int(msg[20:20 + 2], 16) & 0x80 else "off"
                level = int(msg[22:22 + 2], 16) & 0x7F
                modeChange = "on" if int(msg[22:22 + 2], 16) & 0x80 else "off"

                state = level

                if level == 0:
                    state = "off"
                if level == 100:
                    state = "on"

                # readingsBeginUpdate(hash)
                self.update_state(code, "stairwellFunction", stairwellFunction, "1", channel=channel)
                self.update_state(code, "stairwellTime", stairwellTime, "1", channel=channel)
                self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel)
                self.update_state(code, "duskAutomatic", duskAuto, "1", channel=channel)
                self.update_state(code, "sunAutomatic", sunAuto, "1", channel=channel)
                self.update_state(code, "sunMode", sunMode, "1", channel=channel)
                self.update_state(code, "manualMode", manualMode, "1", channel=channel)
                self.update_state(code, "dawnAutomatic", dawnAuto, "1", channel=channel)
                self.update_state(code, "saveIntermediateOnStop", intemedSave, "1", channel=channel)
                self.update_state(code, "runningTime", runningTime, "1", channel=channel)
                self.update_state(code, "intermediateValue", intemedVal, "1", channel=channel)
                self.update_state(code, "intermediateMode", intermedMode, "1", channel=channel)
                self.update_state(code, "level", level, "1", channel=channel)
                self.update_state(code, "modeChange", modeChange, "1", channel=channel)
                self.update_state(code, "state", state, "1", channel=channel)
                # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch

            # Thermostat -- not tested yet
            elif format == "27":  # pragma: no cover
                temperature1 = "%0.1f" % (((int(msg[8:8 + 4], 16) & 0x07FF) - 400) / 10)
                temperature2 = "%0.1f" % (((int(msg[12:12 + 4], 16) & 0x07FF) - 400) / 10)
                tempThreshold1 = "%0.1f" % ((int(msg[16:16 + 2], 16) - 80) / 2)
                tempThreshold2 = "%0.1f" % ((int(msg[18:18 + 2], 16) - 80) / 2)
                tempThreshold3 = "%0.1f" % ((int(msg[20:20 + 2], 16) - 80) / 2)
                tempThreshold4 = "%0.1f" % ((int(msg[22:22 + 2], 16) - 80) / 2)
                desiredTemp = "%0.1f" % ((int(msg[26:26 + 2], 16) - 80) / 2)
                output = "on" if int(msg[8:8 + 2], 16) & 0x08 else "off"
                manualOverride = "on" if int(msg[8:8 + 2], 16) & 0x10 else "off"
                actTempLimit = (int(msg[8:8 + 2], 16) & 0x60) >> 5
                timerAuto = "on" if int(msg[12:12 + 2], 16) & 0x08 else "off"
                manualMode = "on" if int(msg[12:12 + 2], 16) & 0x10 else "off"

                state = "T: temperature1 desired: desiredTemp"

                # readingsBeginUpdate(hash)
                self.update_state(code, "measured-temp", temperature1, "1", channel=channel)
                self.update_state(code, "measured-temp2", temperature2, "1", channel=channel)
                self.update_state(code, "temperatureThreshold1", tempThreshold1, "1", channel=channel)
                self.update_state(code, "temperatureThreshold2", tempThreshold2, "1", channel=channel)
                self.update_state(code, "temperatureThreshold3", tempThreshold3, "1", channel=channel)
                self.update_state(code, "temperatureThreshold4", tempThreshold4, "1", channel=channel)
                self.update_state(code, "desired-temp", desiredTemp, "1", channel=channel)
                self.update_state(code, "output", output, "1", channel=channel)
                self.update_state(code, "manualOverride", manualOverride, "1", channel=channel)
                self.update_state(code, "actTempLimit", actTempLimit, "1", channel=channel)
                self.update_state(code, "timeAutomatic", timerAuto, "1", channel=channel)
                self.update_state(code, "manualMode", manualMode, "1", channel=channel)

                self.update_state(code, "state", state, "1", channel=channel)
                # readingsEndUpdate(hash, 1)

            else:
                logger.warning("DUOFERN unknown msg: {}".format(msg))


        # Wandtaster, Funksender UP, Handsender, Sensoren
        elif msg[0:2] == "0f" and msg[4:6] in ['07', '0e']:  # pragma: no cover
            id = msg[4:4 + 4]

            if id not in sensorMsg:
                logger.warning("unknown message {}".format(msg))

            chan = msg[sensorMsg[id]['chan'] * 2 + 2:sensorMsg[id]['chan'] * 2 + 4]
            if code[0:2] in ("61", "70", "71"):
                chan = "01"

            chans = []
            if (sensorMsg[id]["chan"] == 5):
                chanCount = 4 if (code[0:2] == "73") else 5
                for x in range(0, chanCount):
                    if ((0x01 << x) & int(chan, 16)):
                        chans.append(x + 1)


            else:
                chans.append(chan)

            if code[0:2] in ("65", "69", "74"):
                # module_definition01 = self.modules['by_code'][code + "00"]
                channel = 0
            #if not module_definition01:
                #DoTrigger("global", "UNDEFINED DUOFERN_code_sensor DUOFERN code00")
                #module_definition01 = self.modules['by_code'][code + "00"]

            #if (module_definition01):
            #    hash = module_definition01
            #    channel = 0

            for chan in chans:
                if id[2:4] in ("1a", "18", "19", "01", "02", "03"):
                    if (id[2:4] == "1a") or (id[0:2] == "0e") or (code[0:2] in ("a0", "a2")):
                        self.update_state(code, "state", sensorMsg[id]['state'] + "." + chan, "1", channel=channel)
                    else:
                        self.update_state(code, "state", sensorMsg[id]['state'], "1", channel=channel)

                    self.update_state(code, "channelchan", sensorMsg[id]['name'], "1", channel=channel)
                else:
                    if (code[0:2] not in ("69", "73")) or (id[2:4] in ("11", "12")):
                        chan = ""
                    if code[0:2] in ("65", "a5", "aa", "ab"):
                        self.update_state(code, "state", sensorMsg[id]['state'], "1", channel=channel)

                    self.update_state(code, "event", sensorMsg[id]['name'] + "." + chan, "1", channel=channel)
                    # DoTrigger(hash["name"], sensorMsg[id][name] + "." + chan)



        # Umweltsensor Wetter -- not tested yet
        elif msg[0:8] == "0f011322":  # pragma: no cover
            # module_definition01 = self.modules['by_code'][code + "00"]
            # if not module_definition01:
            #     DoTrigger("global", "UNDEFINED DUOFERN_code_sensor DUOFERN code00")
            #     module_definition01 = self.modules['by_code'][code + "00"]
            #
            # hash = module_definition01
            channel = 0

            brightnessExp = 1000 if int(msg[8:8 + 4], 16) & 0x0400 else 1
            brightness = (int(msg[8:8 + 4], 16) & 0x01FF) * brightnessExp
            sunDirection = int(msg[14:14 + 2], 16) * 1.5
            sunHeight = int(msg[16:16 + 2], 16) - 90
            temperature = ((int(msg[18:18 + 4], 16) & 0x7FFF) - 400) / 10
            isRaining = 1 if int(msg[18:18 + 4], 16) & 0x8000 else 0
            wind = (int(msg[22:22 + 4], 16) & 0x03FF) / 10

            state = "T: {}".format(temperature)
            state += " W: {}".format(wind)
            state += " IR: ".format(isRaining)
            state += " B: ".format(brightness)

            # readingsBeginUpdate(hash)
            self.update_state(code, "brightness", brightness, "1", channel=channel)
            self.update_state(code, "sunDirection", sunDirection, "1", channel=channel)
            self.update_state(code, "sunHeight", sunHeight, "1", channel=channel)
            self.update_state(code, "temperature", temperature, "1", channel=channel)
            self.update_state(code, "isRaining", isRaining, "1", channel=channel)
            self.update_state(code, "state", state, "1", channel=channel)
            self.update_state(code, "wind", wind, "1", channel=channel)
            # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch

        # Umweltsensor Zeit
        elif msg[0:8] == "0fff1020":  # pragma: no cover
            # module_definition01 = self.modules['by_code'][code + "00"]
            # if (not module_definition01):
            #     DoTrigger("global", "UNDEFINED DUOFERN_code_sensor DUOFERN code00")
            #     module_definition01 = self.modules['by_code'][code + "00"]
            #
            # hash = module_definition01
            channel = 0

            year = msg[12:12 + 2]
            month = msg[14:14 + 2]
            day = msg[18:18 + 2]
            hour = msg[20:20 + 2]
            minute = msg[22:22 + 2]
            second = msg[24:24 + 2]

            # readingsBeginUpdate(hash)
            self.update_state(code, "date", "20" + str(year) + "-" + str(month) + "-" + str(day), "1", channel=channel)
            self.update_state(code, "time", str(hour) + ":" + str(minute) + ":" + str(second), "1", channel=channel)
            # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch

        # Umweltsensor Konfiguration
        elif msg[0:7] == "0fff1b2" and msg[7] in ["0", "1", "2", "3", "4", "5", "6", "7", "8"]:  # pragma: no cover
            reg = msg[6:6 + 2] - 21
            regVal = msg[8:8 + 20]

            # module_definition01 = self.modules['by_code'][code + "00"]
            # if not module_definition01:
            #     DoTrigger("global", "UNDEFINED DUOFERN_code_sensor DUOFERN {}00".format(code))
            #     module_definition01 = self.modules['by_code'][code + "00"]

            # hash = module_definition01
            channel = 0

            logger.warning("Weather sensor not supported yet")
            #del hash['READINGS']['configModified']
            self.update_state(code, ".regreg", "regVal", "1", channel=channel)
            # self.update_state(code, "regreg", "regVal", "1", channel=channel)

            #DUOFERN_DecodeWeatherSensorConfig(hash)

            # Rauchmelder Batterie
        elif msg[0:8] == "0fff1323":  # pragma: no cover
            battery = "low" if int(msg[8:8 + 2], 16) <= 10 else "ok"
            batteryLevel = int(msg[8:8 + 2], 16)

            # readingsBeginUpdate(hash)
            self.update_state(code, "battery", battery, "1", channel=channel)
            self.update_state(code, "batteryLevel", batteryLevel, "1", channel=channel)
            # readingsEndUpdate(hash, 1)  # Notify is done by Dispatch

            # ACK, Befehl vom Aktor empfangen
        elif msg[0:8] == "810003cc":
            logger.debug("ack received {}".format(self.modules['by_code'][code]))
            #hash['helper']['timeout']['t'] = hash['name']["timeout"]["60"]
            ##InternalTimer(gettimeofday()+hash['helper']['timeout']{t}, "DUOFERN_StatusTimeout", hash, 0)
            #hash['helper']['timeout']['count'] = 4

        # NACK, Befehl nicht vom Aktor empfangen
        elif msg[0:8] == "810108aa":
            logger.warning("missing ack for {}".format(self.modules['by_code'][code]))
            # self.update_state(code, "state", "MISSING ACK", "1", channel=channel)
            # foreach (grep (/^channel_/, keys%{hash})){
            #   chnHash = module_definitions{hash->{_}}
            #   readingsSingleUpdate(chnHash, "state", "MISSING ACK", 1)
            # }
            # Log3 hash, 3, "DUOFERN error: name MISSING ACK"

        else:
            logger.warning("Unknown msg: {}".format(msg))

#        if module_definition01:
#            DoTrigger(module_definition01['name'], None)
#        if module_definition02:
#            DoTrigger(module_definition02['name'], None)

        return name

    def send(self, cmd):
        self.send_hook(cmd)

    def set(self, code, cmd, *args, channel: int = None):
        # my (hash, @a) = @_
        # b = @a

        # return "set name needs at least one parameter" if(@a < 2)

        #        me     = shift @a
        #        cmd    = shift @a
        arg = args[0] if len(args) >= 1 else None
        arg2 = args[1] if len(args) > 1 else None
        assert len(code) == 6, "code should be 6 hex digits"
        # code = code[0:0 + 6]
        name = self.modules['by_code'][code]['name']

        # sets

        if code[0:2] == "49":
            sets = merge_dicts(setsBasic, setsDefaultRollerShutter, setsRolloTube)
        if code[0:2] in ("42", "4b", "4c", "70"):
            sets = merge_dicts(setsBasic, setsDefaultRollerShutter, setsTroll, {"blindsMode:on,off": ""})
        if code[0:2] == "47":
            sets = merge_dicts(setsBasic, setsDefaultRollerShutter, setsTroll)
        if code[0:2] in ("40", "41", "61"):
            sets = merge_dicts(setsBasic, setsDefaultRollerShutter)  # if (code =~ /^(40|41|61)..../)
        if code[0:2] == "69":
            sets = merge_dicts(setsBasic, setsUmweltsensor)  # if (code =~ /^69..../)
        if code[0:2] == "69" and len(code) >= 8 and code[6:8] == "00":
            sets = merge_dicts(setsUmweltsensor00)  # if (code =~ /^69....00/)
        if code[0:2] == "69" and len(code) >= 8 and code[6:8] == "01":
            sets = merge_dicts(setsDefaultRollerShutter, setsUmweltsensor01)  # if (code =~ /^69....01/)
        if code[0:2] == "43" and len(code) >= 8 and code[6:8] in ("01", "02"):
            sets = merge_dicts(*setsSwitchActor)  # if (code =~ /^43....(01|02)/)
        if code[0:2] in ("43", "65", "74"):
            sets = merge_dicts(setsBasic, {"getStatus:noArg": ""})  # if (code =~ /^(43|65|74)..../)
        if code[0:2] in ("46", "71"):
            sets = merge_dicts(setsBasic, setsSwitchActor)  # if (code =~ /^(46|71)..../)
        if code[0:2] == "4e":
            sets = merge_dicts(setsBasic, setsSX5)  # if (code =~ /^4E..../)
        if code[0:2] == "48":
            sets = merge_dicts(setsBasic, setsDimmer)  # if (code =~ /^48..../)
        if code[0:2] == "73":
            sets = merge_dicts(setsBasic, setsThermostat)  # if (code =~ /^73..../)
        if code[0:2] in ("65", "74") and len(code) >= 8 and code[6:8] == "01":
            sets = merge_dicts(setsSwitchActor)  # if (code =~ /^(65|74)....01/)

        blindsMode = "off" if not "blindsMode" in self.modules['by_code'][code] else self.modules['by_code'][code]
        if (blindsMode == "on"):
            sets = merge_dicts(sets, setsBlinds)

        logger.debug(sets.keys())  # join(" ", sort keys sets)
        if cmd in commandsStatus:
            buf = duoStatusRequest
            buf = buf.replace("nn", commandsStatus[cmd])
            buf = buf.replace("yyyyyy", code)
            self.send(buf)
            return None

        elif cmd == "clear":
            keys = self.modules['by_code'][code].keys()
            for key in keys:
                if key != 'name':
                    self.modules['by_code'][code].__delitem__(key)
            return None
            # cH = (hash)
            # delete _->{READINGS} foreach (@cH)
            # return undef

        elif cmd == "getConfig":
            buf = duoWeatherConfig
            buf = buf.replace("yyyyyy", code)
            self.send(buf)
            return None

        elif cmd == "writeConfig":
            for x in range(0, 8):
                # for(my x=0; x<8; x++)    {
                regV = "00000000000000000000" if not ".reg{}".format(x) in self.modules['by_code'][code] else \
                    self.modules['by_code'][code][".reg{}".format(x)]
                reg = "%02x" % (x + 0x81)
                buf = duoWeatherWriteConfig
                buf = buf.replace("yyyyyy", code)
                buf = buf.replace("rr", reg)
                buf = buf.replace("nnnnnnnnnnnnnnnnnnnn", regV)
                self.send(buf)

            if "configModified" in self.modules['by_code'][code]:
                self.modules['by_code'][code].__delitem__("configModified")

            # delete hash->{READINGS}{configModified}
            return None

        elif cmd == "time":
            buf = duoSetTime

            # my (sec,min,hour,mday,month,year,wday,yday,isdst) = localtime

            year, month, mday, hour, min, sec, wday, yday, isdst, = time.localtime()

            wday = wday - 1 if wday != 0 else 7  # wday = (wday==0 ? 7 : wday-1)
            m = "%02d%02d%02d%02d" % (year - 100, month + 1, wday, mday)
            n = "%02d%02d%02d" % (hour, min, sec)

            buf = buf.replace("mmmmmmmm", m)
            buf = buf.replace("nnnnnn", n)
            buf = buf.replace("yyyyyy", code)
            self.send(buf)
            return None

        elif cmd in wCmds:
            logger.error("this has not been implemented yet")
            # if code[0:2] =="69" and len(code)>=8 and code[6:8] == "00":
            #     return "This command is not allowed for this device."
            # regs=[]
            # if len(args)<1:
            #     return "Missing argument"
            #
            # local_args = args + ("off","off","off","off")
            #
            # for x in range(0,8):#for(my x=0; x<8; x++)    {
            #     temp="00000000000000000000" if not ".regx" in self.modules['by_code'][code] else self.modules['by_code'][code][".regx"]
            #     regs.append(temp)
            #
            #
            # if cmd ==  "triggerSun":
            #     logger.error("this needs to be implemented (triggerSun)")
            #     # newargs=[]
            #     # for _arg in args:
            #     #     if (_arg != "off"):
            #     #         args2 = _arg.split(":")
            #     #         temp = _arg
            #     #         #return "Wrong argument _" if (args2[0] !~ m/^\d+/ || args2[0] < 1 || args2[0] > 100)
            #     #         #return "Wrong argument _" if (args2[1] !~ m/^\d+/ || args2[1] < 1 || args2[1] > 30)
            #     #         #return "Wrong argument _" if (args2[2] !~ m/^\d+/ || args2[2] < 1 || args2[2] > 30)
            #     #
            #     #         if (len(args2) < 3):
            #     #             return "Missing argument"
            #     #         if (int(args2[0]) < 1 or args2[0] > 100) or\
            #     #              (int(args2[1]) < 1 or args2[1] > 30) or\
            #     #              (int(args2[2]) < 1 or args2[2] > 30):
            #     #             return "Wrong argument {}".format(_arg)
            #     #         _arg = ((args2[0]-1)<<12) | ((args2[1]-1)<<19) | ((args2[2]-1)<<24)
            #     #
            #     #         if(len(args2) > 3):
            #     #             if (int(args2[3]) < -5 or int(args2[3]) > 26):
            #     #                 return "Wrong argument {}".format(temp)
            #     #             _arg |= (((int(args2[3]) + 5) << 7) | 0x40)
            #     #     newargs.append(arg)
            #     # args=newargs
            #
            #
            #
            #
            #
            # if cmd ==  "triggerSunDirection":
            #     logger.error("not implemented (triggersundirection)")
            #     # for(my x=0; x<5; x++)    {
            #     #     if (args[x] ne "off") {
            #     #         args2 = split(/:/, args[x])
            #     #         return "Missing argument" if(@args2 < 2)
            #     #         return "Wrong argument args[x]" if (args2[0] !~ m/^\d+(\.\d+|)/ || args2[0] < 0 || args2[0] > 315)
            #     #         return "Wrong argument args[x]" if (args2[1] !~ m/^\d+/ || args2[1] < 45 || args2[1] > 180)
            #     #         args2[0] = int((args2[0]+11.25)/22.5)
            #     #         args2[1] = int((args2[1]+22.5)/45)
            #     #         args2[0] = 15 - (args2[1]*2) if ((args2[0] + args2[1]*2) > 15)
            #     #         args[x] = (args2[0]+args2[1]) | ((args2[1])<<4) | 0x80
            #     #     } else {
            #     #         tSunHeight = map{hex(_)} unpack 'x66A2x8A2x8A2x8A2x8A2', regs
            #     #         if (tSunHeight[x] & 0x18) {
            #     #             args[x] = 0x81
            #     #         } else {
            #     #             args[x] = 0x01
            #     #         }
            #     #     }
            #     # }
            # }
            #
            # if cmd ==  "triggerSunHeight":
            #     logger.error("not implemented (triggersundirection)")
            #     #
            #     #
            #     # for(my x=0; x<5; x++)    {
            #     #     if (args[x] ne "off") {
            #     #         args2 = split(/:/, args[x])
            #     #         return "Missing argument" if(@args2 < 2)
            #     #         return "Wrong argument1 args[x]" if (args2[0] !~ m/^\d+/ || args2[0] < 0 || args2[0] > 90)
            #     #         return "Wrong argument2 args[x]" if (args2[1] !~ m/^\d+/ || args2[1] < 20 || args2[1] > 60)
            #     #         args2[0] = int((args2[0]+6.5)/13)
            #     #         args2[1] = int((args2[1]+13)/26)
            #     #         args2[0] = 7 - (args2[1]*2) if ((args2[0] + args2[1]*2) > 7)
            #     #         args[x] = ((args2[0]+args2[1])<<8) | ((args2[1])<<11) | 0x80
            #     #     } else {
            #     #         tSunDir = map{hex(_)} unpack 'x68A2x8A2x8A2x8A2x8A2', regs
            #     #         if (tSunDir[x] & 0x70) {
            #     #             args[x] = 0x0180
            #     #         } else {
            #     #             args[x] = 0x0100
            #     #         }
            #     #     }
            #     # }
            # }
            #
            # for (my c = 0; c<wCmds{cmd}{count}; c++) {
            #     pad = 0
            #
            #     if (wCmds{cmd}{size} == 4) {
            #         pad = int(c / 2)*2
            #         pad = c if (cmd =~ m/^triggerSun.*/)
            #     }
            #     regStart = (wCmds{cmd}{reg} * 10 + wCmds{cmd}{byte} + pad + c * wCmds{cmd}{size} )*2
            #
            #     reg = hex(substr(regs, regStart, wCmds{cmd}{size} * 2))
            #
            #     if((args[c] =~ m/^[-\d]+/) and (args[c] >=    wCmds{cmd}{min}) and (args[c] <=    wCmds{cmd}{max})) {
            #         reg &= ~(wCmds{cmd}{mask})
            #         reg |= wCmds{cmd}{enable}
            #         reg |= ((args[c] +    wCmds{cmd}{offset})<<wCmds{cmd}{shift}) & wCmds{cmd}{mask}
            #
            #     } elsif ((args[c] == "off") and (wCmds{cmd}{enable} > 0)) {
            #         reg &= ~(wCmds{cmd}{enable})
            #
            #     } elsif ((args[c] == "on") and (wCmds{cmd}{min} == 0) and (wCmds{cmd}{max} == 0)) {
            #         reg |= wCmds{cmd}{enable}
            #
            #     } else {
            #         return "wrong argument ".args[c]
            #
            #     }
            #
            #     size = wCmds{cmd}{size}*2
            #
            #     substr(regs, regStart ,size, sprintf("0".size."x",reg))
            #
            # }
            #
            # @regsA = unpack('(A20)*', regs)
            #
            # # readingsBeginUpdate(hash)
            # for(my x=0; x<8; x++)    {
            #     readingsBulkUpdate(hash, ".regx", regsA[x], 0)
            #     #readingsBulkUpdate(hash, "regx", regsA[x], 0)
            # }
            # readingsBulkUpdate(hash, "configModified", 1, 0)
            # # readingsEndUpdate(hash, 1)
            #
            # DUOFERN_DecodeWeatherSensorConfig(hash)
            # return undef

        elif cmd in commands:
            logger.info("command valid")
            subCmd = None
            if channel is None:
                chanNo = "01"
            else:
                chanNo = "{:02x}".format(channel)
            argV = "00"
            argW = "0000"
            timer = "00"
            buf = duoCommand
            command = None

            if 'noArg' in commands[cmd]:
                if (arg and (arg == "timer")):
                    timer = "01"
                subCmd = "noArg"
                argV = "00"

            elif 'value' in commands[cmd]:
                if (arg2 and (arg2 == "timer")):
                    timer = "01"
                if arg is None:
                    return "Missing argument"
                if (int(arg) < 0 or int(arg) > 100):
                    raise Exception("Wrong argument arg")
                subCmd = "value"
                argV = "%02x" % arg

            elif 'value2' in commands[cmd]:
                if arg is None:
                    return "Missing argument"
                if int(arg) < 0 or int(arg) > 3200:
                    raise Exception("Wrong argument arg")
                subCmd = "value2"
                argW = "%04x" % (arg * 10)

            elif 'value3' in commands[cmd]:
                maxArg = 150
                if code[0:2] == "48":
                    maxArg = 255
                if arg2 and (arg2 == "timer"):
                    timer = "01"
                if arg is None:
                    return "Missing argument"
                if int(arg) < 0 or int(arg) > maxArg:
                    raise Exception("Wrong argument arg")
                subCmd = "value3"
                argV = "%02x" % arg

            elif 'value4' in commands[cmd]:
                if arg2 and (arg2 == "timer"):
                    timer = "01"
                if arg is None:
                    return "Missing argument"
                if int(arg) < 0 or int(arg) > 5000:
                    raise Exception("Wrong argument arg")
                arg = arg / 100
                subCmd = "value4"
                argV = "%02x" % arg

            elif 'temp1' in commands[cmd]:
                if arg is None:
                    return "Missing Argument"
                if int(arg) < -40 or int(arg) > 80:
                    return "Wrong argument {}".format(arg)

                # return "Missing argument" if (!defined(arg))
                # return "Wrong argument arg" if (arg !~ m/^\d+(\.\d+|)/ || arg < -40 || arg > 80)
                subCmd = "temp1"
                argW = "%04x" % ((arg * 10) + 400)

            elif 'temp2' in commands[cmd]:
                if arg is None:
                    return "Missing Argument"
                if int(arg) < -40 or int(arg) > 80:
                    return "Wrong argument {}".format(arg)
                subCmd = "temp2"
                argV = "%02x" % ((arg * 2) + 80)

            else:
                if arg is None:
                    return "Missing Argument"
                if (arg2 and (arg2 == "timer")):
                    timer = "01"
                subCmd = arg
                argV = "00"

            if subCmd not in commands[cmd]:
                raise Exception("Wrong argument {}, {}".format(arg, subCmd))

            channel_suffix = ""
            if channel is not None:
                channel_suffix = "_" + chanNo

            position = -1 if "position"+channel_suffix not in self.modules['by_code'][code] else \
                self.modules['by_code'][code]["position"+channel_suffix]
            # toggleUpDown    = AttrVal(name, "toggleUpDown", "0")
            toggleUpDown = self.modules['by_code'][code]['toggleUpDown'+channel_suffix] \
                if 'toggleUpDown'+channel_suffix in self.modules['by_code'][code] else 0
            moving = "stop" if "moving"+channel_suffix not in self.modules['by_code'][code] else \
                self.modules['by_code'][code]["moving"+channel_suffix]
            timeAutomatic = "on" if "timeAutomatic"+channel_suffix not in self.modules['by_code'][code] else \
                self.modules['by_code'][code]["timeAutomatic"+channel_suffix]
            dawnAutomatic = "on" if "dawnAutomatic"+channel_suffix not in self.modules['by_code'][code] else \
                self.modules['by_code'][code]["dawnAutomatic"+channel_suffix]
            duskAutomatic = "on" if "duskAutomatic"+channel_suffix not in self.modules['by_code'][code] else \
                self.modules['by_code'][code]["duskAutomatic"+channel_suffix]

            if moving != "stop":
                if cmd in ('up', 'down', 'toggle'):
                    if toggleUpDown:
                        cmd = "stop"
            # self.update_state(code,"moving","moving", channel=channel)

            if ((cmd == "toggle") and (position > -1)):
                self.update_state(code, "moving", "moving", 1, channel=channel)
            if ((cmd == "dawn") and (dawnAutomatic == "on") and (position > 0)):
                self.update_state(code, "moving", "up", 1, channel=channel)
            if ((cmd == "dusk") and (duskAutomatic == "on") and (position < 100) and (position > -1)):
                self.update_state(code, "moving", "down", 1, channel=channel)

            if timer == "00" or timeAutomatic == "on":
                if ((cmd == "up") and (position > 0)):
                    self.update_state(code, "moving", "up", 1, channel=channel)
                if ((cmd == "down") and (position < 100) and (position > -1)):
                    self.update_state(code, "moving", "down", 1, channel=channel)

            if cmd == "position":
                if arg > position:
                    self.update_state(code, "moving", "down", 1, channel=channel)
                elif (arg < position):
                    self.update_state(code, "moving", "up", 1, channel=channel)
                else:
                    self.update_state(code, "moving", "stop", 1, channel=channel)

            command = commands[cmd][subCmd]

            buf = buf.replace("yyyyyy", code)
            buf = buf.replace("nnnnnnnnnnnnnnnnnnnn", command)
            buf = buf.replace("nn", argV)
            buf = buf.replace("tt", timer)
            buf = buf.replace("wwww", argW)
            buf = buf.replace("kk", chanNo)
            logger.debug("trying to send {}".format(buf))
            self.send(buf)
            #            if ('device' in self.modules['by_code'][code]):
            # hash = defs{hash->{device}}

        else:
            raise Exception("command {} not found".format(cmd))


# return SetExtensions(hash, list, @b)

if __name__ == "__main__":
    formatter = logging.Formatter('%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)



#!/usr/bin/env python3
# coding=utf-8
#   python interface for dufoern usb stick
#   Copyright (C) 2017 Paul Görgen
#   Rough python re-write of the FHEM duofern modules by telekatz, also licensed under GPLv2
#   This re-write contains only negligible amounts of original code
#   apart from some comments to facilitate translation of the not-yet
#   translated parts of the original software. Modification dates are
#   documented as submits to the git repository of this code, currently
#   maintained at https://github.com/gluap/pyduofern.git

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import asyncio
import logging

import serial_asyncio

from pyduofern.duofern_stick import DuofernStickAsync

logging.basicConfig(level=logging.DEBUG)
loop = asyncio.get_event_loop()

coro = serial_asyncio.create_serial_connection(loop, lambda: DuofernStickAsync(loop), '/dev/ttyUSB0', baudrate=115200)
f, proto = loop.run_until_complete(coro)
# proto.handshake()

initialization = asyncio.ensure_future(proto.handshake())
asyncio.wait(initialization)


def cb(a):
    logging.info(a)
    asyncio.ensure_future(proto.command("409882", "position", 10))


proto.available.add_done_callback(cb)


loop.run_forever()



# coding=utf-8
#   python interface for dufoern usb stick
#   Copyright (C) 2017 Paul Görgen
#   Rough python re-write of the FHEM duofern modules by telekatz, also licensed under GPLv2
#   This re-write contains only negligible amounts of original code
#   apart from some comments to facilitate translation of the not-yet
#   translated parts of the original software. Modification dates are
#   documented as submits to the git repository of this code, currently
#   maintained at https://github.com/gluap/pyduofern.git

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import asyncio
import codecs
import json
import logging
import os
import os.path
import tempfile
import threading
import time

import serial
import serial.tools.list_ports

from .duofern import Duofern
from .exceptions import DuofernTimeoutException, DuofernException


def hex(stuff):
    return codecs.getencoder('hex')(stuff)[0].decode("utf-8")


logger = logging.getLogger(__file__)

duoInit1 = "01000000000000000000000000000000000000000000"
duoInit2 = "0E000000000000000000000000000000000000000000"
duoSetDongle = "0Azzzzzz000100000000000000000000000000000000"
duoInit3 = "14140000000000000000000000000000000000000000"
duoSetPairs = "03nnyyyyyy0000000000000000000000000000000000"
duoInitEnd = "10010000000000000000000000000000000000000000"
duoACK = "81000000000000000000000000000000000000000000"
duoStatusRequest = "0DFF0F400000000000000000000000000000FFFFFF01"
duoStartPair = "04000000000000000000000000000000000000000000"
duoStopPair = "05000000000000000000000000000000000000000000"
duoStartUnpair = "07000000000000000000000000000000000000000000"
duoStopUnpair = "08000000000000000000000000000000000000000000"
duoRemotePair = "0D0006010000000000000000000000000000yyyyyy01"


def refresh_serial_connection(function):
    def new_funtion(*args, **kwargs):
        self = args[0]
        if self.serial_connection.isOpen():
            return function(*args, **kwargs)
        else:  # pragma: no cover
            self.serial_connection.open()
            return function(*args, **kwargs)

    return new_funtion


class DuofernStick(object):
    def __init__(self, system_code=None, config_file_json=None, duofern_parser=None, recording=None,
                 changes_callback=None, ephemeral=None, *args, **kwargs):
        """
        :param device: path to com port opened by usb stick (e.g. /dev/ttyUSB0)
        :param system_code: system code
        :param config_file_json: path to config file. use the same one to conveniently update info about your system
        :param duofern_parser: parser object. Unless you hacked your own one just leave None and it
         defaults to pyduofern.duofern.Duofern()
        """
        super().__init__(*args, **kwargs)
        self.config_file = None
        self.ephemeral = ephemeral
        config_file_json = self.prepare_config(config_file_json)
        self.config_file = config_file_json

        if duofern_parser is None:
            duofern_parser = Duofern(send_hook=self.add_serial_and_send, changes_callback=changes_callback)

        self.duofern_parser = duofern_parser
        self.running = False
        self.pairing = False
        self.unpairing = False

        self.system_code = None
        if system_code is not None:
            if not ephemeral and 'system_code' in self.config:
                assert self.config['system_code'].lower() == system_code.lower(), \
                    'System code passed as argument "{}" differs from config file "{}", please manually change the ' \
                    'config file {} if this is what you intended. If you change the code you paired your devices with' \
                    ' you might have to reset them and re-pair.'.format(system_code,
                                                                        self.config['system_code'],
                                                                        os.path.abspath(config_file_json))

            self.system_code = system_code
        elif 'system_code' in self.config and not ephemeral:  # pragma: no cover
            self.system_code = self.config['system_code']
        else:
            raise DuofernException("No system code specified. Since the system code is a security feature no default"
                                   "can be provided. Please re-run wiht a valid system code {}".format(self.config))
        assert len(self.system_code) == 4, "system code (serial) must be a string of 4 hexadecimal numbers"

        self.pairing = False
        self.unpairing = False
        self.write_queue = []
        if not ephemeral:
            self.config['system_code'] = self.system_code
        self._dump_config()
        self.initialized = False

        if recording is None and 'recording' in self.config:
            recording = bool(self.config['recording'])

        self.recording = recording
        if recording:
            self._initialize_recording()

    def prepare_config(self, config_file_json):
        if config_file_json is None:  # pragma: no cover
            config_file_json = os.path.expanduser("~/.duofern.json")
        if os.path.isfile(config_file_json):
            try:
                with open(config_file_json, "r") as config_file_fh:
                    self.config = json.load(config_file_fh)
            except json.decoder.JSONDecodeError:  # pragma: no cover
                self.config = {'devices': []}
                logger.info('failed reading config')
        else:
            logger.info('config is not file')
            self.config = {'devices': []}
        return config_file_json

    def _initialize_recording(self):
        if 'recording_dir' in self.config:
            dir = self.config['recording_dir']
            assert not os.path.isfile(dir), 'must pass existing or creatable dir as `recording_dir` in config'
            if not os.path.isdir(dir):
                os.makedirs(dir)
            record_filename = os.path.join(dir, "duofern_record_{}".format(str(time.time())))
        else:
            record_filename = tempfile.mktemp(prefix="duofern_record_")
            print("recording to  {}".format(record_filename))
        self.record_filename = record_filename

    def _initialize(self, **kwargs):  # pragma: no cover
        raise NotImplementedError("need to use an implementation of the Duofernstick")

    def _simple_write(self, **kwargs):  # pragma: no cover
        raise NotImplementedError("need to use an implementation of the Duofernstick")

    def send(self, msg, **kwargs):  # pragma: no cover
        raise NotImplementedError("need to use an implementation of the Duofernstick")

    def add_updates_callback(self, callback):
        self.duofern_parser.changes_callback = callback

    def _dump_config(self):
        with open(self.config_file, "w") as config_fh:
            json.dump(self.config, config_fh, indent=4)

    def process_message(self, message):
        if self.recording:
            with open(self.record_filename, "a") as recorder:
                recorder.write("received {}\n".format(message))
                recorder.flush()
        if message[0:2] == '81':
            #            logger.debug("got Acknowledged")
            # return
            # self.handle_write_queue()
            return
        if message[0:4] == '0602':
            logger.info("got pairing reply")
            self.pairing = False
            self.duofern_parser.parse(message)
            self.sync_devices()
            return
        # if ($rmsg =~ m / 0602.{40} / ) {
        #    my %addvals = (RAWMSG => $rmsg);
        #    Dispatch($hash, $rmsg, \%addvals) if ($hash->{pair});
        #    delete($hash->{pair});
        #    RemoveInternalTimer($hash);
        #    return undef;
        #
        elif message[0:4] == '0603':
            logger.info("got unpairing reply")
            self.unpairing = False
            self.duofern_parser.parse(message)
            self.sync_devices()
            return
        # } elsif ($rmsg =~ m/0603.{40}/) {
        #    my %addvals = (RAWMSG => $rmsg);
        #    Dispatch($hash, $rmsg, \%addvals) if ($hash->{unpair});
        #    delete($hash->{unpair});
        #    RemoveInternalTimer($hash);
        #    return undef;
        #
        elif message[0:6] == '0FFF11':
            return

        elif message[0:8] == '81000000':
            return
            #  } elsif ($rmsg =~ m/0FFF11.{38}/) {
            #    return undef;
            #
            #  } elsif ($rmsg =~ m/81000000.{36}/) {
            #    return undef;
            #
            #  }
            #
            #  my %addvals = (RAWMSG => $rmsg);
            #  Dispatch($hash, $rmsg, \%addvals);
        #        logger.info("got {}".format(message))
        self.duofern_parser.parse(message)

    def clean_config(self):
        self.config = dict(devices=[])
        self._dump_config()

    def sync_devices(self):
        known_codes = [device['id'].lower() for device in self.config['devices']]
        logger.debug("known codes {}".format(known_codes))
        for module_id in self.duofern_parser.modules['by_code']:
            if module_id.lower() not in known_codes:
                self.config['devices'].append({'id': module_id, 'name': module_id})
            logger.info("paired new device {}".format(module_id))
        self._dump_config()

    def set_name(self, id, name):
        logger.info("renaming device {} to {}".format(id, name))
        self.config['devices'] = [device for device in self.config['devices'] if device['id'].lower() != id.lower()]
        self.config['devices'].append({'id': id, 'name': name})
        self._dump_config()
        self._initialize()

    def handle_write_queue(self):
        if len(self.write_queue) > 0:
            tosend = self.write_queue.pop()
            logger.info("sending {} from write queue, {} msgs left in queue".format(tosend, len(self.write_queue)))
            self._simple_write(tosend)

    def stop_pair(self):
        self.send(duoStopPair)
        self.pairing = False

    def stop_unpair(self):
        self.send(duoStopUnpair)
        self.unpairing = False

    def pair(self, timeout=10):
        self.send(duoStartPair)
        threading.Timer(timeout, self.stop_pair).start()
        self.pairing = True

    def unpair(self, timeout=10):
        self.send(duoStartUnpair)
        threading.Timer(10, self.stop_unpair).start()
        self.unpairing = True

    def remote(self, code, timeout=10):
        self.send(duoRemotePair.replace('yyyyyy', code))
        threading.Timer(timeout, self.stop_pair).start()
        self.pairing = True

    def test_callback(self, arg):
        self.duofern_parser.parse(arg)


def one_time_callback(protocol, _message, name, future):
    logger.info("{} answer for {}".format(_message, name))
    if not future.cancelled():
        future.set_result(_message)
        future.done()
        protocol.callback = None


async def send_and_await_reply(protocol, message, message_identifier):
    future = asyncio.Future()
    protocol.callback = lambda message: one_time_callback(protocol, message, message_identifier, future)
    protocol.send(message)
    try:
        result = await future
        logger.info("got reply {}".format(result))
    except asyncio.CancelledError:
        logger.info("future was cancelled waiting for reply")


class DuofernStickAsync(DuofernStick, asyncio.Protocol):
    def __init__(self, loop=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.duofern_parser.asyncio = True
        self.initialization_step = 0
        self.loop = loop
        self.write_queue = asyncio.Queue()
        self._ready = asyncio.Event()
        self.transport = None
        self.buffer = bytearray(b'')

        self.last_packet = 0.0
        self.callback = None

        if loop == None:
            loop = asyncio.get_event_loop()

        self.send_loop = asyncio.ensure_future(self._send_messages(), loop=loop)

        self.available = asyncio.Future()

        # DuofernStick.__init__(self, device, system_code, config_file_json, duofern_parser)

    #        self.serial_connection = serial.Serial(self.port, baudrate=115200, timeout=1)
    #        self.running = False

    def command(self, *args, **kwargs):
        if self.recording:
            with open(self.record_filename, "a") as recorder:
                recorder.write("sending_command {} {}\n".format(args, kwargs))
                recorder.flush()
        self.duofern_parser.set(*args, **kwargs)

    async def add_serial_and_send(self, msg):
        message = msg.replace("zzzzzz", "6f" + self.system_code)
        logger.info("sending {}".format(message))
        self.send(message)
        logger.info("added {} to write queue".format(message))

    def connection_made(self, transport):
        self.transport = transport
        logger.info('port opened {}')
        transport.serial.rts = False
        self.buffer = bytearray(b'')
        self.last_packet = time.time()
        self._ready.set()

    def data_received(self, data):
        if self.last_packet + 0.05 < time.time() and not hasattr(self.transport, 'unittesting'):
            self.buffer = bytearray(b'')
        self.last_packet = time.time()
        self.buffer += bytearray(data)
        while len(self.buffer) >= 22:
            if self.recording:
                with open(self.record_filename, "a") as recorder:
                    recorder.write("received {}\n".format(hex(self.buffer[0:22])))
                    recorder.flush()
            if not hex(self.buffer[0:22]) == duoACK:
                self.send(duoACK)
            if hasattr(self, 'callback') and self.callback is not None:
                self.callback(hex(self.buffer[0:22]))
            elif self.initialized:
                self.process_message(hex(self.buffer[0:22]))
            self.buffer = self.buffer[22:]

    def pause_writing(self):  # pragma: no cover
        logger.info('pause writing')
        logger.info(self.transport.get_write_buffer_size())

    def resume_writing(self):  # pragma: no cover
        logger.info(self.transport.get_write_buffer_size())
        logger.info('resume writing')

    def parse(self, packet):
        logger.info(packet)

    def send(self, data, **kwargs):
        """ Feed a message to the sender coroutine. """
        tosend = bytearray.fromhex(data)
        if self.recording:
            with open(self.record_filename, "a") as recorder:
                recorder.write("sent {}\n".format(data))
                recorder.flush()
        self.write_queue.put_nowait(tosend)

    async def _send_messages(self):
        """ Send messages to the server as they become available. """
        await self._ready.wait()
        logger.debug("Starting async send loop!")
        while True:
            try:
                logger.info("sending from stack")
                data = await self.write_queue.get()
                self.transport.write(data)
            except asyncio.CancelledError:
                logger.info("Got CancelledError, stopping send loop")
                break
            except Exception as exc:
                raise

    def parse_regular(self, packet):
        logger.info(packet)

    async def handshake(self):
        if not hasattr(self.transport, 'unittesting'):
            await asyncio.sleep(2)
            logger.info("now handshaking")
        await send_and_await_reply(self, duoInit1, "init 1")
        await send_and_await_reply(self, duoInit2, "init 2")
        await send_and_await_reply(self, duoSetDongle.replace("zzzzzz", "6f" + self.system_code), "SetDongle")
        self.send(duoACK)
        await send_and_await_reply(self, duoInit3, "init 3")
        self.send(duoACK)
        if 'devices' in self.config and self.config['devices']:
            counter = 0
            for device in self.config['devices']:
                # devices with id other than 6 characters
                # were previously sub-devices of another device with 6 characters
                # (i.e. devices representing a single channel)
                # but are no longer relevant
                if len(device['id']) != 6:
                    continue
                hex_to_write = duoSetPairs.replace('nn', '{:02X}'.format(counter)).replace('yyyyyy', device['id'])
                await send_and_await_reply(self, hex_to_write, "SetPairs")
                self.send(duoACK)
                counter += 1
                self.duofern_parser.add_device(device['id'], device['name'])

        await send_and_await_reply(self, duoInitEnd, "duoInitEnd")
        self.send(duoACK)
        await send_and_await_reply(self, duoStatusRequest, "duoInitEnd")
        self.send(duoACK)
        self.available.set_result(True)
        self.initialized = True


class DuofernStickThreaded(DuofernStick, threading.Thread):
    def __init__(self, serial_port=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if serial_port is None:
            try:
                self.port = serial.tools.list_ports.comports()[0].device
            except IndexError:
                raise DuofernException(
                    "No serial port configured and unable to autodetect device. Did you plug in your stick?")
            logger.debug("no serial port set, autodetected {} for duofern".format(self.port))
        else:
            self.port = serial_port

        # DuofernStick.__init__(self, device, system_code, config_file_json, duofern_parser)
        self.serial_connection = serial.Serial(self.port, baudrate=115200, timeout=1)
        self.running = False

    def _read_answer(self, some_string):  # ReadAnswer
        """read an answer..."""
        logger.debug("should read {}".format(some_string))
        self.serial_connection.timeout = 1
        response = bytearray(self.serial_connection.read(22))
        if len(response) < 22:
            raise DuofernTimeoutException
        logger.debug("response {}".format(hex(response)))

        if self.recording:
            with open(self.record_filename, "a") as recorder:
                recorder.write("received {}\n".format(hex(response)))

        return hex(response)

    def _initialize(self):  # DoInit
        for i in range(0, 4):
            self._simple_write(duoInit1)
            try:
                self._read_answer("INIT1")
            except DuofernTimeoutException:  # pragma: no cover
                continue

            self._simple_write(duoInit2)
            try:
                self._read_answer("INIT2")
            except DuofernTimeoutException:  # pragma: no cover
                continue

            buf = duoSetDongle.replace("zzzzzz", "6f" + self.system_code)
            self._simple_write(buf)
            try:
                self._read_answer("SetDongle")
            except DuofernTimeoutException:  # pragma: no cover
                continue

            self._simple_write(duoACK)
            self._simple_write(duoInit3)
            try:
                self._read_answer("INIT3")

            except DuofernTimeoutException:  # pragma: no cover
                continue
            self._simple_write(duoACK)

            if "devices" in self.config:
                counter = 0
                for device in self.config['devices']:
                    # devices with id other than 6 characters
                    # were previously sub-devices of another device with 6 characters
                    # (i.e. devices representing a single channel)
                    # but are no longer relevant
                    if len(device['id']) != 6:
                        continue

                    hex_to_write = duoSetPairs.replace('nn', '{:02X}'.format(counter)).replace('yyyyyy', device['id'])
                    self._simple_write(hex_to_write)
                    try:
                        self._read_answer("SetPairs")
                    except DuofernTimeoutException:  # pragma: no cover
                        continue
                    self._simple_write(duoACK)
                    counter += 1
                    self.duofern_parser.add_device(device['id'], device['name'])

            # my counter = 0
            # foreach (@pairs){
            #   buf = duoSetPairs
            #   my chex .= sprintf "%02x", counter
            #   buf =~ s/nn/chex/
            #   buf =~ s/yyyyyy/_/
            #   self._simple_write(buf)
            #   (err, buf) = self._read_answer("SetPairs")
            #   next if(err)
            #   self._simple_write(duoACK)
            #   counter++
            # }

            self._simple_write(duoInitEnd)
            try:
                self._read_answer("INIT3")  # look @ original?
            except DuofernTimeoutException:  # pragma: no cover
                return False
            self._simple_write(duoACK)

            self._simple_write(duoStatusRequest)
            try:
                self._read_answer("statusRequest")
            except DuofernTimeoutException:  # pragma: no cover
                continue
            self._simple_write(duoACK)

            # readingsSingleUpdate(hash, "state", "Initialized", 1)
            return True

        raise DuofernTimeoutException("Initialization failed ")

    # DUOFERNSTICK_SimpleWrite(@)
    @refresh_serial_connection
    def _simple_write(self, string_to_write):  # SimpleWrite
        """Just write data"""
        logger.debug("writing  {}".format(string_to_write))
        hex_to_write = string_to_write.replace(" ", '')
        if self.recording:
            with open(self.record_filename, "a") as recorder:
                recorder.write("sent {}\n".format(hex_to_write))

        data_to_write = bytearray.fromhex(hex_to_write)
        if not self.serial_connection.isOpen():
            self.serial_connection.open()
        self.serial_connection.write(data_to_write)

    def command(self, *args, **kwargs):
        if self.recording:
            with open(self.record_filename, "a") as recorder:
                recorder.write("sending_command {} {}\n".format(args,kwargs))
        self.duofern_parser.set(*args, **kwargs)

    def add_serial_and_send(self, msg):
        message = msg.replace("zzzzzz", "6f" + self.system_code)
        logger.debug("sending {}".format(message))
        self.send(message)
        logger.debug("added {} to write queue".format(message))

    def run(self):
        self.running = True
        self._initialize()
        while self.running:
            self.serial_connection.timeout = .05
            if not self.serial_connection.isOpen():
                self.serial_connection.open()
            try:
                in_data = hex(self.serial_connection.read(22))
            except TypeError:
                continue
            if len(in_data) == 44:
                if in_data != duoACK:
                    self._simple_write(duoACK)
                try:
                    self.process_message(in_data)
                except Exception as exc:
                    logger.exception(exc)
            self.serial_connection.timeout = 1
            if len(self.write_queue) > 0:
                self.handle_write_queue()

    def stop(self):
        self.running = False
        self.serial_connection.close()

    def pair(self, timeout=10):
        super(DuofernStickThreaded, self).pair(timeout)
        threading.Timer(timeout, self.stop_pair).start()

    def unpair(self, timeout=10):
        super(DuofernStickThreaded, self).unpair(timeout)
        threading.Timer(timeout, self.stop_unpair).start()

    def send(self, msg, **kwargs):
        logger.debug("sending {}".format(msg))
        self.write_queue.append(msg)
        logger.debug("added {} to write queue".format(msg))
        return



# coding=utf-8
from setuptools import setup

from pyduofern import __version__ as version

setup(name='pyduofern',
      version=version,
      description='Library for controlling Rademacher DuoFern actors using python. Requires the Rademacher'
                  'Duofern USB Stick Art.-Nr.: 70000093',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          'Programming Language :: Python :: 3.4',
      ],

      url='https://github.com/gluap/pyduofern',
      author='Paul Görgen',
      author_email='pypi@pgoergen.de',
      license='GPL-2.0',

      packages=['pyduofern'],

      install_requires=[
          'pyserial', 'pyserial-asyncio'
      ],

      zip_safe=False,
      include_package_data=False,

      tests_require=[
          'tox', 'pytest', 'pytest-asyncio'
      ],

      scripts=["scripts/duofern_cli.py"],

      long_description=open('README.rst', 'r').read()
      )
